"""Functional interface to hypergraph methods and assorted utilities."""

from collections import Counter

from scipy.special import comb

from ..exception import XGIError

__all__ = [
    "num_edges_order",
    "max_edge_order",
    "is_uniform",
    "is_possible_order",
    "edge_neighborhood",
    "degree_counts",
    "degree_histogram",
    "unique_edge_sizes",
    "density",
    "incidence_density",
]


def num_edges_order(H, d=None):
    """The number of edges of order d.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest.

    d : int, optional
        The order of edges to count. If None (default), counts
        for all orders.

    Returns
    -------
    int
        The number of edges of order d

    See Also
    --------
    max_edge_order
    """

    if d is not None:
        return len(H.edges.filterby("order", d))
    else:
        return H.num_edges


def max_edge_order(H):
    """The maximum order of edges in the hypergraph.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest.

    Returns
    -------
    int
        Maximum order of edges in hypergraph.

    See Also
    --------
    num_edges_order
    """
    if H._edge:
        d_max = max(len(edge) for edge in H._edge.values()) - 1
    else:
        d_max = 0 if H._node else None
    return d_max


def is_possible_order(H, d):
    """Whether the specified order is between 0 (singletons) and the maximum order.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest.
    d : int
        Order for which to check.

    Returns
    -------
    bool
        Whether `d` is a possible order.

    """
    d_max = max_edge_order(H)
    return (d >= 0) and (d <= d_max)


def is_uniform(H):
    """Order of uniformity if the hypergraph is uniform, or False.

    A hypergraph is uniform if all its edges have the same order.

    Returns d if the hypergraph is d-uniform, that is if all edges
    in the hypergraph (excluding singletons) have the same degree d.
    Returns False if not uniform.

    Returns
    -------
    d : int or False
        If the hypergraph is d-uniform, return d, or False otherwise.

    Examples
    --------
    This function can be used as a boolean check:

    >>> import xgi
    >>> H = xgi.Hypergraph([(0, 1, 2), (1, 2, 3), (2, 3, 4)])
    >>> xgi.is_uniform(H)
    2
    >>> if xgi.is_uniform(H): print('H is uniform!')
    H is uniform!

    """
    edge_sizes = unique_edge_sizes(H)
    if 1 in edge_sizes:
        edge_sizes.remove(1)  # discard singleton edges

    if not edge_sizes or len(edge_sizes) != 1:
        return False

    return edge_sizes.pop() - 1  # order of all edges


def edge_neighborhood(H, n, include_self=False):
    """The edge neighborhood of the specified node.

    The edge neighborhood of a node `n` in a hypergraph `H` is an edgelist of all the
    edges containing `n` and its edges are all the edges in `H` that contain `n`.
    Usually, the edge neighborhood does not include `n` itself.  This can be controlled
    with `include_self`.

    Parameters
    ----------
    H : Hypergraph
        THe hypergraph of interest
    n : node
        Node whose edge_neighborhood is needed.
    include_self : bool, optional
        Whether the edge_neighborhood contains `n`.
        By default, False.
    Returns
    -------
    list
        An edgelist of the edge_neighborhood of `n`.

    See Also
    --------
    neighbors

    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph([[1, 2, 3], [3, 4], [4, 5, 6]])
    >>> H.nodes.neighbors(3)
    {1, 2, 4}
    >>> xgi.edge_neighborhood(H, 3)
    [{1, 2}, {4}]
    >>> xgi.edge_neighborhood(H, 3, include_self=True)
    [{1, 2, 3}, {3, 4}]

    """
    if include_self:
        return [H.edges.members(e) for e in H.nodes.memberships(n)]
    else:
        return [H.edges.members(e) - {n} for e in H.nodes.memberships(n)]


def degree_counts(H, order=None):
    """Returns a list of the the number of occurrences of each degree value.

    The counts correspond to degrees from 0 to max(degree).

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest
    order: int, optional
        Order of edges to take into account. If None (default),
        consider all edges.

    Returns
    -------
    list
        A list of frequencies of degrees.
        The degree values are the index in the list.

    Notes
    -----
    Note: the bins are width one, hence len(list) can be large
    (Order(num_edges))

    The degree is defined as the number of edges to which a node belongs.
    A node belonging only to a singleton edge will thus have degree 1 and
    contribute accordingly to the degree count.

    Examples
    --------
    >>> import xgi
    >>> hyperedge_list = [[1, 2], [2, 3, 4]]
    >>> H = xgi.Hypergraph(hyperedge_list)
    >>> xgi.degree_counts(H)
    [0, 3, 1]

    """
    counts = Counter(H.degree(order=order).values())
    return [counts.get(i, 0) for i in range(max(counts) + 1)]


def degree_histogram(H):
    """Returns a degree histogram including bin centers (degree values).

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    tuple of lists
        First entry is observed degrees (bin centers),
            second entry is degree count (histogram height)
    Notes
    -----
    Note: the bins are width one, hence there will be an entry
    for every observed degree.

    Examples
    --------
    >>> import xgi
    >>> hyperedge_list = [[1, 2], [2, 3, 4]]
    >>> H = xgi.Hypergraph(hyperedge_list)
    >>> xgi.degree_histogram(H)
    ([1, 2], [3, 1])

    """
    counts = Counter(H.degree().values())
    degrees = []
    heights = []
    for d, c in sorted(counts.items(), key=lambda kv: kv[0]):
        degrees.append(d)
        heights.append(c)
    return degrees, heights


def unique_edge_sizes(H):
    """A function that returns the unique edge sizes.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    list()
        The unique edge sizes in ascending order by size.
    """
    return sorted(set(H.edges.size.asnumpy()))


def density(H, order=None, max_order=None, ignore_singletons=False):
    r"""Hypergraph density.

    The density of a hypergraph is the number of existing edges divided by the number of
    possible edges.

    Let `H` have :math:`n` nodes and :math:`m` hyperedges. Then,

    * `density(H) =` :math:`\frac{m}{2^n - 1}`,
    * `density(H, ignore_singletons=True) =` :math:`\frac{m}{2^n - 1 - n}`.

    Here, :math:`2^n` is the total possible number of hyperedges on `H`, from which we
    subtract :math:`1` because the empty hyperedge is not considered.  We subtract an
    additional :math:`n` when singletons are not considered.

    Now assume `H` has :math:`a` edges with order :math:`1` and :math:`b` edges with
    order :math:`2`.  Then,

    * `density(H, order=1) =` :math:`\frac{a}{{n \choose 2}}`,
    * `density(H, order=2) =` :math:`\frac{b}{{n \choose 3}}`,
    * `density(H, max_order=1) =` :math:`\frac{a}{{n \choose 1} + {n \choose 2}}`,
    * `density(H, max_order=1, ignore_singletons=True) =` :math:`\frac{a}{{n \choose 2}}`,
    * `density(H, max_order=2) =` :math:`\frac{m}{{n \choose 1} + {n \choose 2} + {n \choose 3}}`,
    * `density(H, max_order=2, ignore_singletons=True) =` :math:`\frac{m}{{n \choose 2} + {n \choose 3}}`,

    Parameters
    ---------
    order : int, optional
        If not None, only count edges of the specified order.
        By default, None.

    max_order : int, optional
        If not None, only count edges of order up to this value, inclusive.
        By default, None.

    ignore_singletons : bool, optional
        Whether to consider singleton edges.  Ignored if `order` is not None and
        different from :math:`0`. By default, False.

    See Also
    --------
    :func:`incidence_density`

    Notes
    -----
    If both `order` and `max_order` are not None, `max_order` is ignored.

    """
    n = H.num_nodes
    if n < 1:
        raise XGIError("Density not defined for empty hypergraph")
    if H.num_edges < 1:
        return 0.0

    def order_filter(val, mode):
        return H.edges.filterby("order", val, mode=mode)

    if order is None and max_order is None:
        numer = H.num_edges
        denom = 2**n - 1
        if ignore_singletons:
            numer -= len(order_filter(0, mode="eq"))
            denom -= n

    elif order is None and max_order is not None:
        if max_order >= n:
            raise ValueError("max_order must be smaller than the number of nodes")
        numer = len(order_filter(max_order, "leq"))
        denom = sum(comb(n, _ord + 1, exact=True) for _ord in range(max_order + 1))
        if ignore_singletons:
            numer -= len(order_filter(0, mode="eq"))
            denom -= n

    elif order is not None:  # ignore max_order
        if order >= n:
            raise ValueError("order must be smaller than the number of nodes")
        if ignore_singletons and order == 0:
            return 0.0
        numer = len(order_filter(order, mode="eq"))
        denom = comb(n, order + 1, exact=True)

    try:
        return float(numer) / denom
    except ZeroDivisionError:
        return 0.0


def incidence_density(H, order=None, max_order=None, ignore_singletons=False):
    r"""Density of the incidence matrix.

    The incidence matrix of a hypergraph contains one row per node and one column per
    edge.  An entry is non-zero when the corresponding node is a member of the
    corresponding edge.  The density of this matrix is the number of non-zero entries
    divided by the total number of entries.

    Parameters
    ---------
    order : int, optional
        If not None, only count edges of the specified order.
        By default, None.

    max_order : int, optional
        If not None, only count edges of order up to this value, inclusive.
        By default, None.

    ignore_singletons : bool, optional
        Whether to consider singleton edges.  Ignored if `order` is not None and
        different from :math:`0`. By default, False.

    See Also
    --------
    :func:`density`

    Notes
    -----
    If both `order` and `max_order` are not None, `max_order` is ignored.

    The parameters `order`, `max_order` and `ignore_singletons` have a similar effect on
    the denominator as they have in :func:`density`.

    """
    n = H.num_nodes
    if n < 1:
        raise XGIError("Density not defined for empty hypergraph")
    if H.num_edges < 1:
        return 0.0

    edges_to_count = H.edges
    if order is None and max_order is None:
        if ignore_singletons:
            edges_to_count = edges_to_count.filterby("order", 0, "gt")

    elif order is None and max_order is not None:
        if max_order >= n:
            raise ValueError("max_order must be smaller than the number of nodes")
        edges_to_count = edges_to_count.filterby("order", max_order, "leq")
        if ignore_singletons:
            edges_to_count = edges_to_count.filterby("order", 0, "gt")

    elif order is not None:  # ignore max_order
        if order >= n:
            raise ValueError("order must be smaller than the number of nodes")
        if ignore_singletons and order == 0:
            return 0.0
        edges_to_count = edges_to_count.filterby("order", order, "eq")

    denom = n * len(edges_to_count)
    numer = edges_to_count.size.asnumpy().sum()  # size, not order
    try:
        # cast both to float because sometimes 0 / 0.0 may issue a warning instead of
        # raising ZeroDivisionError
        return float(numer) / float(denom)
    except ZeroDivisionError:
        return 0.0



from . import assortativity, centrality, clustering, connected, properties
from .assortativity import *
from .centrality import *
from .clustering import *
from .connected import *
from .shortest_path import *
from .properties import *



"""Generate random (non-uniform) hypergraphs."""
import math
import random
import warnings
from collections import defaultdict
from itertools import combinations

import numpy as np
from scipy.special import comb

from .classic import empty_hypergraph
from .lattice import ring_lattice

__all__ = [
    "random_hypergraph",
    "chung_lu_hypergraph",
    "dcsbm_hypergraph",
    "watts_strogatz_hypergraph",
]


def random_hypergraph(N, ps, order=None, seed=None):
    """Generates a random hypergraph

    Generate N nodes, and connect any d+1 nodes
    by a hyperedge with probability ps[d-1].

    Parameters
    ----------
    N : int
        Number of nodes
    ps : list of float
        List of probabilities (between 0 and 1) to create a
        hyperedge at each order d between any d+1 nodes. For example,
        ps[0] is the wiring probability of any edge (2 nodes), ps[1]
        of any triangles (3 nodes).
    order: int of None (default)
        If None, ignore. If int, generates a uniform hypergraph with edges
        of order `order` (ps must have only one element).
    seed : integer or None (default)
            Seed for the random number generator.

    Returns
    -------
    Hypergraph object
        The generated hypergraph

    References
    ----------
    Described as 'random hypergraph' by M. Dewar et al. in https://arxiv.org/abs/1703.07686

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01])

    """
    if seed is not None:
        np.random.seed(seed)

    if order is not None:
        if len(ps) != 1:
            raise ValueError("ps must contain a single element if order is an int")

    if (np.any(np.array(ps) < 0)) or (np.any(np.array(ps) > 1)):
        raise ValueError("All elements of ps must be between 0 and 1 included.")

    nodes = range(N)
    hyperedges = []

    for i, p in enumerate(ps):

        if order is not None:
            d = order
        else:
            d = i + 1  # order, ps[0] is prob of edges (d=1)

        potential_edges = combinations(nodes, d + 1)
        n_comb = comb(N, d + 1, exact=True)
        mask = np.random.random(size=n_comb) <= p  # True if edge to keep

        edges_to_add = [e for e, val in zip(potential_edges, mask) if val]

        hyperedges += edges_to_add

    H = empty_hypergraph()
    H.add_nodes_from(nodes)
    H.add_edges_from(hyperedges)

    return H


def chung_lu_hypergraph(k1, k2, seed=None):
    """A function to generate a Chung-Lu hypergraph

    Parameters
    ----------
    k1 : dictionary
        Dictionary where the keys are node ids
        and the values are node degrees.
    k2 : dictionary
        Dictionary where the keys are edge ids
        and the values are edge sizes.
    seed : integer or None (default)
            The seed for the random number generator.

    Returns
    -------
    Hypergraph object
        The generated hypergraph

    Warns
    -----
    warnings.warn
        If the sums of the edge sizes and node degrees are not equal, the
        algorithm still runs, but raises a warning.

    Notes
    -----
    The sums of k1 and k2 should be the same. If they are not the same,
    this function returns a warning but still runs.

    References
    ----------
    Implemented by Mirah Shi in HyperNetX and described for
    bipartite networks by Aksoy et al. in https://doi.org/10.1093/comnet/cnx001

    Example
    -------
    >>> import xgi
    >>> import random
    >>> n = 100
    >>> k1 = {i : random.randint(1, 100) for i in range(n)}
    >>> k2 = {i : sorted(k1.values())[i] for i in range(n)}
    >>> H = xgi.chung_lu_hypergraph(k1, k2)

    """
    if seed is not None:
        random.seed(seed)

    # sort dictionary by degree in decreasing order
    node_labels = [n for n, _ in sorted(k1.items(), key=lambda d: d[1], reverse=True)]
    edge_labels = [m for m, _ in sorted(k2.items(), key=lambda d: d[1], reverse=True)]

    m = len(k2)

    if sum(k1.values()) != sum(k2.values()):
        warnings.warn(
            "The sum of the degree sequence does not match the sum of the size sequence"
        )

    S = sum(k1.values())

    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    for u in node_labels:
        j = 0
        v = edge_labels[j]  # start from beginning every time
        p = min((k1[u] * k2[v]) / S, 1)

        while j < m:
            if p != 1:
                r = random.random()
                try:
                    j = j + math.floor(math.log(r) / math.log(1 - p))
                except ZeroDivisionError:
                    j = np.inf

            if j < m:
                v = edge_labels[j]
                q = min((k1[u] * k2[v]) / S, 1)
                r = random.random()
                if r < q / p:
                    # no duplicates
                    H.add_node_to_edge(v, u)
                p = q
                j = j + 1

    return H


def dcsbm_hypergraph(k1, k2, g1, g2, omega, seed=None):
    """A function to generate a Degree-Corrected Stochastic Block Model
    (DCSBM) hypergraph.

    Parameters
    ----------
    k1 : dict
        This is a dictionary where the keys are node ids
        and the values are node degrees.
    k2 : dict
        This is a dictionary where the keys are edge ids
        and the values are edge sizes.
    g1 : dict
        This a dictionary where the keys are node ids
        and the values are the group ids to which the node belongs.
        The keys must match the keys of k1.
    g2 : dict
        This a dictionary where the keys are edge ids
        and the values are the group ids to which the edge belongs.
        The keys must match the keys of k2.
    omega : 2D numpy array
        This is a matrix with entries which specify the number of edges
        between a given node community and edge community.
        The number of rows must match the number of node communities
        and the number of columns must match the number of edge
        communities.
    seed : int or None (default)
        Seed for the random number generator.

    Returns
    -------
    Hypergraph

    Warns
    -----
    warnings.warn
        If the sums of the edge sizes and node degrees are not equal, the
        algorithm still runs, but raises a warning.
        Also if the sum of the omega matrix does not match the sum of degrees,
        a warning is raised.

    Notes
    -----
    The sums of k1 and k2 should be the same. If they are not the same, this function
    returns a warning but still runs. The sum of k1 (and k2) and omega should be the
    same. If they are not the same, this function returns a warning but still runs and
    the number of entries in the incidence matrix is determined by the omega matrix.

    References
    ----------
    Implemented by Mirah Shi in HyperNetX and described for bipartite networks by
    Larremore et al. in https://doi.org/10.1103/PhysRevE.90.012805

    Examples
    --------
    >>> import xgi; import random; import numpy as np
    >>> n = 50
    >>> k1 = {i : random.randint(1, n) for i in range(n)}
    >>> k2 = {i : sorted(k1.values())[i] for i in range(n)}
    >>> g1 = {i : random.choice([0, 1]) for i in range(n)}
    >>> g2 = {i : random.choice([0, 1]) for i in range(n)}
    >>> omega = np.array([[n//2, 10], [10, n//2]])
    >>> # H = xgi.dcsbm_hypergraph(k1, k2, g1, g2, omega)

    """
    if seed is not None:
        random.seed(seed)

    # sort dictionary by degree in decreasing order
    node_labels = [n for n, _ in sorted(k1.items(), key=lambda d: d[1], reverse=True)]
    edge_labels = [m for m, _ in sorted(k2.items(), key=lambda d: d[1], reverse=True)]

    # Verify that the sum of node and edge degrees and the sum of node degrees and the
    # sum of community connection matrix differ by less than a single edge.
    if abs(sum(k1.values()) - sum(k2.values())) > 1:
        warnings.warn(
            "The sum of the degree sequence does not match the sum of the size sequence"
        )

    if abs(sum(k1.values()) - np.sum(omega)) > 1:
        warnings.warn(
            "The sum of the degree sequence does not "
            "match the entries in the omega matrix"
        )

    # get indices for each community
    community1_nodes = defaultdict(list)
    for label in node_labels:
        group = g1[label]
        community1_nodes[group].append(label)

    community2_nodes = defaultdict(list)
    for label in edge_labels:
        group = g2[label]
        community2_nodes[group].append(label)

    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    kappa1 = defaultdict(lambda: 0)
    kappa2 = defaultdict(lambda: 0)
    for id, g in g1.items():
        kappa1[g] += k1[id]
    for id, g in g2.items():
        kappa2[g] += k2[id]

    for group1 in community1_nodes.keys():
        for group2 in community2_nodes.keys():
            # for each constant probability patch
            try:
                group_constant = omega[group1, group2] / (
                    kappa1[group1] * kappa2[group2]
                )
            except ZeroDivisionError:
                group_constant = 0

            for u in community1_nodes[group1]:
                j = 0
                v = community2_nodes[group2][j]  # start from beginning every time
                # max probability
                p = min(k1[u] * k2[v] * group_constant, 1)
                while j < len(community2_nodes[group2]):
                    if p != 1:
                        r = random.random()
                        try:
                            j = j + math.floor(math.log(r) / math.log(1 - p))
                        except ZeroDivisionError:
                            j = np.inf
                    if j < len(community2_nodes[group2]):
                        v = community2_nodes[group2][j]
                        q = min((k1[u] * k2[v]) * group_constant, 1)
                        r = random.random()
                        if r < q / p:
                            # no duplicates
                            H.add_node_to_edge(v, u)
                        p = q
                        j = j + 1
    return H


def watts_strogatz_hypergraph(n, d, k, l, p, seed=None):
    if seed is not None:
        np.random.seed(seed)
    H = ring_lattice(n, d, k, l)
    to_remove = []
    to_add = []
    for e in H.edges:
        if np.random.random() < p:
            to_remove.append(e)
            node = min(H.edges.members(e))
            neighbors = np.random.choice(H.nodes, size=d - 1)
            to_add.append(np.append(neighbors, node))
    H.remove_edges_from(to_remove)
    H.add_edges_from(to_add)
    return H



"""Load a data set from the xgi-data repository or a local file."""
import json
import os
from warnings import warn

from .. import convert
from ..exception import XGIError
from ..utils import request_json_from_url, request_json_from_url_cached

__all__ = ["load_xgi_data", "download_xgi_data"]


def load_xgi_data(
    dataset=None,
    cache=True,
    read=False,
    path="",
    nodetype=None,
    edgetype=None,
    max_order=None,
):
    """Load a data set from the xgi-data repository or a local file.

    Parameters
    ----------
    dataset : str, optional
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository. If None (default), prints
        the list of available datasets.
    cache : bool, optional
        Whether to cache the input data, by default True.
    read : bool, optional
        If read==True, search for a local copy of the data set. Use the local
        copy if it exists, otherwise use the xgi-data repository.
        By default, False.
    path : str, optional
        Path to a local copy of the data set
    nodetype : type, optional
        Type to cast the node ID to, by default None.
    edgetype : type, optional
        Type to cast the edge ID to, by default None.
    max_order: int, optional
        Maximum order of edges to add to the hypergraph, by default None.

    Returns
    -------
    Hypergraph
        The loaded hypergraph.

    Raises
    ------
    XGIError
       The specified dataset does not exist.
    """
    index_url = (
        "https://gitlab.com/complexgroupinteractions/"
        "xgi-data/-/raw/main/index.json?inline=false"
    )

    # If no dataset is specified, print a list of the available datasets.
    if dataset is None:

        index_data = request_json_from_url(index_url)
        print("Available datasets are the following:")
        print(*index_data, sep="\n")
        return

    if read:
        cfp = os.path.join(path, dataset + ".json")
        if os.path.exists(cfp):
            data = json.load(open(cfp, "r"))

            return convert.dict_to_hypergraph(
                data, nodetype=nodetype, edgetype=edgetype, max_order=max_order
            )
        else:
            warn(
                f"No local copy was found at {cfp}. The data is requested "
                "from the xgi-data repository instead. To download a local "
                "copy, use `download_xgi_data`."
            )
    data = _request_from_xgi_data(dataset, cache=cache)

    return convert.dict_to_hypergraph(
        data, nodetype=nodetype, edgetype=edgetype, max_order=max_order
    )


def download_xgi_data(dataset, path=""):
    """Make a local copy of a dataset in the xgi-data repository.

    Parameters
    ----------
    dataset : str
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository.

    path : str, optional
        Path to where the local copy should be saved. If none is given, save
        file to local directory.
    """

    jsondata = _request_from_xgi_data(dataset)
    jsonfile = open(os.path.join(path, dataset + ".json"), "w")
    json.dump(jsondata, jsonfile)
    jsonfile.close()


def _request_from_xgi_data(dataset=None, cache=True):
    """Request a dataset from xgi-data.

    Parameters
    ----------
    dataset : str, optional
        Dataset name. Valid options are the top-level tags of the
        index.json file in the xgi-data repository. If None, prints
        the list of available datasets.
    cache : bool, optional
        Whether or not to cache the output

    Returns
    -------
    Data
        The requested data loaded from a json file.

    Raises
    ------
    XGIError
        If the HTTP request is not successful or the dataset does not exist.

    See also
    ---------
    load_xgi_data
    """
    index_url = (
        "https://gitlab.com/complexgroupinteractions/"
        "xgi-data/-/raw/main/index.json?inline=false"
    )

    index_data = request_json_from_url(index_url)

    key = dataset.lower()
    if key not in index_data:
        print("Valid dataset names:")
        print(*index_data, sep="\n")
        raise XGIError("Must choose a valid dataset name!")

    if cache:
        return request_json_from_url_cached(index_data[key]["url"])
    else:
        return request_json_from_url(index_data[key]["url"])



import sys

import setuptools
from setuptools import setup

__version__ = "0.7.4"

if sys.version_info < (3, 8):
    sys.exit("XGI requires Python 3.8 or later.")

name = "xgi"

version = __version__

authors = "XGI Developers"

author_email = "nicholas.landry@uvm.edu"

project_urls = {
    "Documentation": "https://xgi.readthedocs.io/en/latest/",
    "Bug Reports": "https://github.com/xgi-org/xgi/issues",
    "Source": "https://github.com/xgi-org/xgi",
    "PyPI": "https://pypi.org/project/xgi/",
    "GitHub Discussions": "https://github.com/xgi-org/xgi/discussions",
}

description = """XGI is a Python package for higher-order networks."""

with open("long_description.rst") as file:
    long_description = file.read()


def parse_requirements_file(filename):
    with open(filename) as fid:
        requires = [l.strip() for l in fid.readlines() if not l.startswith("#")]
    return requires


extras_require = {
    dep: parse_requirements_file("requirements/" + dep + ".txt")
    for dep in [
        "benchmarks",
        "developer",
        "documentation",
        "release",
        "test",
        "tutorial",
    ]
}

extras_require["all"] = list({item for dep in extras_require.values() for item in dep})

install_requires = parse_requirements_file("requirements/default.txt")

license = "3-Clause BSD license"

setup(
    name=name,
    packages=setuptools.find_packages(),
    version=version,
    author=authors,
    author_email=author_email,
    project_urls=project_urls,
    description=description,
    long_description=long_description,
    install_requires=install_requires,
    extras_require=extras_require,
)



"""Generators for some simplicial complexes.

All the functions in this module return a SimplicialComplex class.

"""

import random
from collections import defaultdict
from itertools import combinations

import networkx as nx
import numpy as np
from scipy.special import comb

from ..core import SimplicialComplex
from ..utils.utilities import find_triangles

__all__ = [
    "random_simplicial_complex",
    "random_flag_complex_d2",
    "random_flag_complex",
    "flag_complex",
    "flag_complex_d2",
]


def random_simplicial_complex(N, ps, seed=None):
    """Generates a random hypergraph

    Generate N nodes, and connect any d+1 nodes
    by a simplex with probability ps[d-1]. For each simplex,
    add all its subfaces if they do not already exist.

    Parameters
    ----------
    N : int
        Number of nodes
    ps : list of float
        List of probabilities (between 0 and 1) to create a
        hyperedge at each order d between any d+1 nodes. For example,
        ps[0] is the wiring probability of any edge (2 nodes), ps[1]
        of any triangles (3 nodes).
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    Simplicialcomplex object
        The generated simplicial complex

    References
    ----------
    Described as 'random simplicial complex' in
    "Simplicial Models of Social Contagion", Nature Communications 10(1), 2485,
    by I. Iacopini, G. Petri, A. Barrat & V. Latora (2019).
    https://doi.org/10.1038/s41467-019-10431-6

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_simplicial_complex(20, [0.1, 0.01])

    """

    if seed is not None:
        np.random.seed(seed)

    if (np.any(np.array(ps) < 0)) or (np.any(np.array(ps) > 1)):
        raise ValueError("All elements of ps must be between 0 and 1 included.")

    nodes = range(N)
    simplices = []

    for i, p in enumerate(ps):
        d = i + 1  # order, ps[0] is prob of edges (d=1)

        potential_simplices = combinations(nodes, d + 1)
        n_comb = comb(N, d + 1, exact=True)
        mask = np.random.random(size=n_comb) <= p  # True if simplex to keep

        simplices_to_add = [e for e, val in zip(potential_simplices, mask) if val]

        simplices += simplices_to_add

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(simplices)

    return S


def flag_complex(G, max_order=2, ps=None, seed=None):
    """Generate a flag (or clique) complex from a
    NetworkX graph by filling all cliques up to dimension max_order.

    Parameters
    ----------
    G : Networkx Graph

    max_order : int
        maximal dimension of simplices to add to the output simplicial complex
    ps: list of float
        List of probabilities (between 0 and 1) to create a
        hyperedge from a clique, at each order d. For example,
        ps[0] is the probability of promoting any 3-node clique (triangle) to
        a 3-hyperedge.
    seed: int or None (default)
        The seed for the random number generator

    Returns
    -------
    S : SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks. `flag_complex_d2`
    is faster to compute up to order 2.

    See also
    --------
    flag_complex_d2

    """
    # This import needs to happen when this function is called, not when it is
    # defined.  Otherwise, a circular import error would happen.
    from ..core import SimplicialComplex

    if seed is not None:
        random.seed(seed)

    nodes = G.nodes()
    N = len(nodes)
    edges = G.edges()

    cliques_to_add = _cliques_to_fill(G, max_order)

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(edges)
    if not ps:  # promote all cliques
        S.add_simplices_from(cliques_to_add, max_order=max_order)
        return S

    # store max cliques per order
    cliques_d = defaultdict(list)
    for x in cliques_to_add:
        cliques_d[len(x)].append(x)

    # promote cliques with a given probability
    for i, p in enumerate(ps[: max_order - 1]):
        d = i + 2  # simplex order
        cliques_d_to_add = [el for el in cliques_d[d + 1] if random.random() <= p]
        S.add_simplices_from(cliques_d_to_add, max_order=max_order)

    return S


def flag_complex_d2(G, p2=None, seed=None):
    """Generate a flag (or clique) complex from a
    NetworkX graph by filling all cliques up to dimension 2.

    Parameters
    ----------
    G : networkx Graph
        Graph to consider
    p2: float
        Probability (between 0 and 1) of filling empty triangles in graph G
    seed: int or None (default)
        The seed for the random number generator

    Returns
    -------
    S : xgi.SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks. This
    is faster than `flag_complex` to compute up to order 2.

    See also
    --------
    flag_complex
    """
    # This import needs to happen when this function is called, not when it is
    # defined.  Otherwise, a circular import error would happen.
    from ..core import SimplicialComplex

    if seed is not None:
        random.seed(seed)

    nodes = G.nodes()
    edges = G.edges()

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(edges)

    triangles_empty = find_triangles(G)

    if p2 is not None:
        triangles = [el for el in triangles_empty if random.random() <= p2]
    else:
        triangles = triangles_empty

    S.add_simplices_from(triangles)

    return S


def random_flag_complex_d2(N, p, seed=None):
    """Generate a maximal simplicial complex (up to order 2) from a :math:`G_{N,p}`
    Erdős-Rényi random graph.

    This proceeds by filling all empty triangles in the graph with 2-simplices.

    Parameters
    ----------
    N : int
        Number of nodes
    p : float
        Probabilities (between 0 and 1) to create an edge
        between any 2 nodes
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks.

    """
    if seed is not None:
        random.seed(seed)

    if (p < 0) or (p > 1):
        raise ValueError("p must be between 0 and 1 included.")

    G = nx.fast_gnp_random_graph(N, p, seed=seed)

    return flag_complex_d2(G)


def random_flag_complex(N, p, max_order=2, seed=None):
    """Generate a flag (or clique) complex from a
    :math:`G_{N,p}` Erdős-Rényi random graph.

    This proceeds by filling all cliques up to dimension max_order.

    Parameters
    ----------
    N : int
        Number of nodes
    p : float
        Probability (between 0 and 1) to create an edge
        between any 2 nodes
    max_order : int
        maximal dimension of simplices to add to the output simplicial complex
    seed : int or None (default)
        The seed for the random number generator

    Returns
    -------
    SimplicialComplex

    Notes
    -----
    Computing all cliques quickly becomes heavy for large networks.

    """
    if (p < 0) or (p > 1):
        raise ValueError("p must be between 0 and 1 included.")

    G = nx.fast_gnp_random_graph(N, p, seed=seed)

    nodes = G.nodes()

    cliques = _cliques_to_fill(G, max_order)

    S = SimplicialComplex()
    S.add_nodes_from(nodes)
    S.add_simplices_from(cliques, max_order=max_order)

    return S


def _cliques_to_fill(G, max_order):
    """Return cliques to fill for flag complexes,
    to be passed to `add_simplices_from`.

    This function was written to speedup flag_complex functions
    by avoiding adding redundant faces.

    Parameters
    ----------
    G : networkx Graph
        Graph to consider
    max_order: int or None
        If None, return maximal cliques. If int, return all cliques
        up to max_order.

    Returns
    -------
    cliques : list
        List of cliques

    """
    if max_order is None:
        cliques = list(nx.find_cliques(G))  # max cliques
    else:  # avoid adding many unnecessary redundant cliques
        cliques = []
        for clique in nx.enumerate_all_cliques(G):  # sorted by size
            if len(clique) == 1:
                continue  # don't add singletons
            if len(clique) <= max_order + 1:
                cliques.append(clique)
            else:
                break  # dont go over whole list if not necessary

    return cliques



"""Load a metabolic network from the BiGG models database."""

from warnings import warn

from ..utils import request_json_from_url, request_json_from_url_cached

__all__ = ["load_bigg_data"]


def load_bigg_data(
    dataset=None,
    cache=True,
):
    """Load a metabolic network from the BiGG models database.

    The Biochemical, Genetic and Genomic (BiGG) knowledge base
    is hosted at http://bigg.ucsd.edu/. It contains metabolic
    reaction networks at the genome scale.

    We represent metabolites as nodes and metabolic reactions
    as directed edges where reactants are the tail of the directed
    edge and the products are the head of the directed edge.

    Parameters
    ----------
    dataset : str, default: None
        Dataset name. Valid options are the "bigg_id" tags in
        http://bigg.ucsd.edu/api/v2/models. If None, prints
        the list of available datasets.
    cache : bool, optional
        Whether to cache the input data

    Returns
    -------
    DiHypergraph
        The loaded dihypergraph.

    Raises
    ------
    XGIError
       The specified dataset does not exist.

    References
    ----------
    Zachary A. King, Justin Lu, Andreas Dräger,
    Philip Miller, Stephen Federowicz, Joshua A. Lerman,
    Ali Ebrahim, Bernhard O. Palsson, Nathan E. Lewis
    Nucleic Acids Research, Volume 44, Issue D1,
    4 January 2016, Pages D515–D522,
    https://doi.org/10.1093/nar/gkv1049
    """

    index_url = "http://bigg.ucsd.edu/api/v2/models"
    base_url = "http://bigg.ucsd.edu/static/models/"

    index_data = request_json_from_url(index_url)

    # If no dataset is specified, print a list of the available datasets.
    if dataset is None:
        ids = []
        for entry in index_data["results"]:
            ids.append(entry["bigg_id"])
        print("Available datasets are the following:")
        print(*ids, sep="\n")
        return

    if cache:
        model_data = request_json_from_url_cached(base_url + dataset + ".json")
    else:
        model_data = request_json_from_url(base_url + dataset + ".json")

    return _bigg_to_dihypergraph(index_data, model_data)


def _bigg_to_dihypergraph(d_index, d_model):
    """Convert a BIGG-formatted dict to dihypergraph.

    Parameters
    ----------
    d : dict
        A BIGG-formatted dict

    Returns
    -------
    DiHypergraph
        The dihypergraph from the selected BIGG model.

    Notes
    -----
    The code for parsing a metabolic reaction is rewritten
    from a function by @pietrotraversa.

    We use the `lower_bound` and `upper_bound` variables to
    determine whether a reaction is a forward, reverse,
    or reversible reaction.
    """
    from .. import DiHypergraph

    DH = DiHypergraph()

    id = d_model["id"]

    DH["name"] = id

    info = next((item for item in d_index["results"] if item["bigg_id"] == id), None)
    DH["organism"] = info["organism"]

    for m in d_model["metabolites"]:
        DH.add_node(m["id"], name=m["name"])

    for r in d_model["reactions"]:
        l = r["lower_bound"]
        u = r["upper_bound"]

        reactants = set()
        products = set()

        # forward direction
        if l >= 0 and u > 0:
            for m, val in r["metabolites"].items():
                if val > 0:
                    products.add(m)
                elif val <= 0:
                    reactants.add(m)

            if not reactants and not products:
                warn(f"{r['id']} is an empty reaction!")
                continue
            DH.add_edge((reactants, products), id=r["id"], name=r["name"])

        # reverse direction
        if l < 0 and u <= 0:
            for m, val in r["metabolites"].items():
                if val >= 0:
                    reactants.add(m)
                elif val < 0:
                    products.add(m)

            if not reactants and not products:
                warn(f"{r['id']} is an empty reaction!")
                continue
            DH.add_edge((reactants, products), id=r["id"], name=r["name"])

        # reversible
        if l < 0 and u > 0:
            for m, val in r["metabolites"].items():
                if val > 0:
                    products.add(m)
                elif val < 0:
                    reactants.add(m)

            if not reactants and not products:
                warn(f"{r['id']} is an empty reaction!")
                continue
            # add forward reaction
            DH.add_edge((reactants, products), id=r["id"], name=r["name"])
            # add reverse reaction
            DH.add_edge(
                (products, reactants), id=str(r["id"]) + "_reverse", name=r["name"]
            )

    return DH



"""General matrices associated to hypergraphs.

Note that the order of the rows and columns of the
matrices in this module correspond to the order in
which nodes/edges are added to the hypergraph or
simplicial complex. If the node and edge IDs are
able to be sorted, the following is an example to sort
by the node and edge IDs.

>>> import xgi
>>> import pandas as pd
>>> H = xgi.Hypergraph([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> I, nodedict, edgedict = xgi.incidence_matrix(H, sparse=False, index=True)
>>> # Sorting the resulting numpy array:
>>> sortedI = I.copy()
>>> sortedI = sortedI[sorted(nodedict, key=nodedict.get), :]
>>> sortedI = sortedI[:, sorted(edgedict, key=edgedict.get)]
>>> sortedI
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])
>>> # Indexing a Pandas dataframe by the node/edge IDs
>>> df = pd.DataFrame(I, index=nodedict.values(), columns=edgedict.values())

If the nodes are already sorted, this order can be preserved by adding the nodes
to the hypergraph prior to adding edges. For example,

>>> import xgi
>>> H = xgi.Hypergraph()
>>> H.add_nodes_from(range(1, 8))
>>> H.add_edges_from([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> xgi.incidence_matrix(H, sparse=False)
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])

"""
from warnings import catch_warnings, warn

import numpy as np
from scipy.sparse import csr_array

__all__ = [
    "incidence_matrix",
    "adjacency_matrix",
    "intersection_profile",
    "degree_matrix",
    "clique_motif_matrix",
]


def incidence_matrix(
    H, order=None, sparse=True, index=False, weight=lambda node, edge, H: 1
):
    """A function to generate a weighted incidence matrix from a Hypergraph object,
    where the rows correspond to nodes and the columns correspond to edges.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    order: int, optional
        Order of interactions to use. If None (default), all orders are used. If int,
        must be >= 1.
    sparse: bool, default: True
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix
    index: bool, default: False
        Specifies whether to output dictionaries mapping the node and edge IDs to
        indices.
    weight: lambda function, default=lambda function outputting 1
        A function specifying the weight, given a node and edge

    Returns
    -------
    I: numpy.ndarray or scipy csr_array
        The incidence matrix, has dimension (n_nodes, n_edges)
    rowdict: dict
        The dictionary mapping indices to node IDs, if index is True
    coldict: dict
        The dictionary mapping indices to edge IDs, if index is True

    """
    node_ids = H.nodes
    edge_ids = H.edges

    if order is not None:
        edge_ids = H.edges.filterby("order", order)
    if not edge_ids or not node_ids:
        if sparse:
            I = csr_array((0, 0), dtype=int)
        else:
            I = np.empty((0, 0), dtype=int)
        return (I, {}, {}) if index else I

    num_edges = len(edge_ids)
    num_nodes = len(node_ids)

    node_dict = dict(zip(node_ids, range(num_nodes)))
    edge_dict = dict(zip(edge_ids, range(num_edges)))

    if index:
        rowdict = {v: k for k, v in node_dict.items()}
        coldict = {v: k for k, v in edge_dict.items()}

    # Compute the non-zero values, row and column indices for the given order
    rows = []
    cols = []
    data = []
    for edge in edge_ids:
        members = H._edge[edge]
        for node in members:
            rows.append(node_dict[node])
            cols.append(edge_dict[edge])
            data.append(weight(node, edge, H))

    # Create the incidence matrix as a CSR matrix
    if sparse:
        I = csr_array((data, (rows, cols)), shape=(num_nodes, num_edges), dtype=int)
    else:
        I = np.zeros((num_nodes, num_edges), dtype=int)
        I[rows, cols] = data

    return (I, rowdict, coldict) if index else I


def adjacency_matrix(H, order=None, sparse=True, s=1, weighted=False, index=False):
    """
    A function to generate an adjacency matrix (N,N) from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    order: int, optional
        Order of interactions to use. If None (default), all orders are used. If int,
        must be >= 1.
    sparse: bool, default: True
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix
    s: int, default: 1
        Specifies the number of overlapping edges to be considered connected.
    weighted: bool
        If True, entry (i, j) [and (j, i)] is the number of edges that connect i and j.
        If False, entry (i, j) [and (j, i] is 1 if i and j share at least one edge
        and 0 otherwise. By default, False.
    index: bool, default: False
        Specifies whether to output disctionaries mapping the node IDs to indices

    Returns
    -------
    if index is True:
        return A, rowdict
    else:
        return A

    Warns
    -----
    warn
        If there are isolated nodes and the matrix is sparse.

    """
    I, rowdict, coldict = incidence_matrix(H, order=order, sparse=sparse, index=True)

    if I.shape == (0, 0):
        if not rowdict:
            A = csr_array((0, 0)) if sparse else np.empty((0, 0))
        if not coldict:
            shape = (H.num_nodes, H.num_nodes)
            A = csr_array(shape, dtype=int) if sparse else np.zeros(shape, dtype=int)
        return (A, {}) if index else A

    A = I.dot(I.T)

    if sparse:
        with catch_warnings(record=True) as w:
            A.setdiag(0)
        if w:
            warn(
                "Forming the adjacency matrix can "
                "be expensive when there are isolated nodes!"
            )
    else:
        np.fill_diagonal(A, 0)

    if not weighted:
        A = (A >= s) * 1
    else:
        A = (A >= s) * A

    if sparse:
        A.eliminate_zeros()

    return (A, rowdict) if index else A


def intersection_profile(H, order=None, sparse=True, index=False):
    """
    A function to generate an intersection profile from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    order: int, optional
        Order of interactions to use. If None (default), all orders are used. If int,
        must be >= 1.
    sparse: bool, default: True
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix
    index: bool, default: False
        Specifies whether to output dictionaries mapping the edge IDs to indices

    Returns
    -------
    if index is True:
        return P, rowdict, coldict
    else:
        return P

    """
    eye, _, coldict = incidence_matrix(H, order=order, sparse=sparse, index=True)
    P = eye.T.dot(eye)
    return (P, coldict) if index else P


def degree_matrix(H, order=None, index=False):
    """Returns the degree of each node as an array

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    order: int, optional
        Order of interactions to use. If None (default), all orders are used. If int,
        must be >= 1.
    index: bool, default: False
        Specifies whether to output disctionaries mapping the node and edge IDs to
        indices.

    Returns
    -------
    if index is True:
        return K, rowdict
    else:
        return K

    """
    eye, rowdict, _ = incidence_matrix(H, order=order, index=True)

    if eye.shape == (0, 0):
        K = np.zeros(H.num_nodes)
    else:
        K = np.ravel(np.sum(eye, axis=1))  # flatten

    return (K, rowdict) if index else K


def clique_motif_matrix(H, sparse=True, index=False):
    """
    A function to generate a weighted clique motif matrix
    from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    sparse: bool, default: True
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix
    index: bool, default: False
        Specifies whether to output dictionaries
        mapping the node and edge IDs to indices

    Returns
    -------
    if index is True:
        return W, rowdict
    else:
        return W

    References
    ----------
    "Higher-order organization of complex networks"
    by Austin Benson, David Gleich, and Jure Leskovic
    https://doi.org/10.1126/science.aad9029

    """
    W, rowdict = adjacency_matrix(H, sparse=sparse, weighted=True, index=True)

    return (W, rowdict) if index else W



class XGIException(Exception):
    """Base class for exceptions in XGI."""


class XGIError(XGIException):
    """Exception for a serious error in XGI"""


class IDNotFound(KeyError):
    """Raised when a node or edge is not in the hypergraph."""


def frozen(*args, **kwargs):
    """Dummy method that raises an error when trying to modify frozen networks

    Raises
    ------
    XGIError
        Raises error when user tries to modify the network

    Examples
    --------
    >>> import xgi
    >>> hyperedge_list = [[1, 2], [2, 3, 4]]
    >>> H = xgi.Hypergraph(hyperedge_list)
    >>> H.freeze()
    >>> H.add_node(5)
    Traceback (most recent call last):
    xgi.exception.XGIError: Frozen higher-order network can't be modified

    """
    raise XGIError("Frozen higher-order network can't be modified")



"""Draw hypergraphs and simplicial complexes with matplotlib."""

from inspect import signature
from itertools import combinations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.patches import FancyArrow
from mpl_toolkits.mplot3d.art3d import Line3DCollection, Poly3DCollection
from networkx import spring_layout
from scipy.spatial import ConvexHull

from .. import convert
from ..algorithms import max_edge_order, unique_edge_sizes
from ..core import DiHypergraph, Hypergraph, SimplicialComplex
from ..exception import XGIError
from .draw_utils import (
    _CCW_sort,
    _color_arg_to_dict,
    _draw_arg_to_arr,
    _draw_init,
    _interp_draw_arg,
    _scalar_arg_to_dict,
    _update_lims,
)
from .layout import _augmented_projection, barycenter_spring_layout

__all__ = [
    "draw",
    "draw_nodes",
    "draw_hyperedges",
    "draw_simplices",
    "draw_node_labels",
    "draw_hyperedge_labels",
    "draw_hypergraph_hull",
    "draw_multilayer",
    "draw_dihypergraph",
]


def draw(
    H,
    pos=None,
    ax=None,
    dyad_color="black",
    dyad_lw=1.5,
    edge_fc=None,
    node_fc="white",
    node_ec="black",
    node_lw=1,
    node_size=15,
    node_shape="o",
    max_order=None,
    node_labels=False,
    hyperedge_labels=False,
    aspect="equal",
    **kwargs,
):
    """Draw hypergraph or simplicial complex.

    Parameters
    ----
    H : Hypergraph or SimplicialComplex.
        Hypergraph to draw
    pos : dict, optional
        If passed, this dictionary of positions node_id:(x,y) is used for placing the
        0-simplices.  If None (default), use the `barycenter_spring_layout` to compute
        the positions.
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    dyad_color : str, dict, iterable, or EdgeStat, optional
        Color of the dyadic links.  If str, use the same color for all edges. If a dict,
        must contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use
        a colormap (specified with dyad_color_cmap) associated to it. By default,
        "black".
    dyad_lw : int, float, dict, iterable, or EdgeStat, optional
        Line width of edges of order 1 (dyadic links).  If int or float, use the same
        width for all edges.  If a dict, must contain (edge_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the edges are
        found in H.edges. If EdgeStat, use a monotonic linear interpolation defined
        between min_dyad_lw and max_dyad_lw. By default, 1.5.
    edge_fc : str, dict, iterable, or EdgeStat, optional
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict,
        must contain (edge_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the hyperedges are found in H.edges. If
        EdgeStat, use the colormap specified with edge_fc_cmap. If None (default), use
        the H.edges.size.
    node_fc : str, dict, iterable, or NodeStat, optional
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use
        the colormap specified with node_fc_cmap. By default, "white".
    node_ec : str, dict, iterable, or NodeStat, optional
        Color of node borders.  If str, use the same color for all nodes.  If a dict,
        must contain (node_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap. By default, "black".
    node_lw : int, float, dict, iterable, or NodeStat, optional
        Line width of the node borders in pixels.  If int or float, use the same width
        for all node borders.  If a dict, must contain (node_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the nodes are
        found in H.nodes. If NodeStat, use a monotonic linear interpolation defined
        between min_node_lw and max_node_lw. By default, 1.
    node_size : int, float, dict, iterable, or NodeStat, optional
        Radius of the nodes in pixels.  If int or float, use the same radius for all
        nodes.  If a dict, must contain (node_id: radius) pairs.  If iterable, assume
        the radiuses are specified in the same order as the nodes are found in
        H.nodes. If NodeStat, use a monotonic linear interpolation defined between
        min_node_size and max_node_size. By default, 15.
    node_shape :  string, optional
        The shape of the node. Specification is as matplotlib.scatter
        marker. Default is "o".
    max_order : int, optional
        Maximum of hyperedges to plot. If None (default), plots all orders.
    node_labels : bool or dict, optional
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
        By default, False.
    hyperedge_labels : bool or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  By default, False.
    aspect : {"auto", "equal"} or float, optional
        Set the aspect ratio of the axes scaling, i.e. y/x-scale. `aspect` is passed
        directly to matplotlib's `ax.set_aspect()`. Default is `equal`. See full
        Set the aspect ratio of the axes scaling, i.e. y/x-scale. `aspect` is passed
        directly to matplotlib's `ax.set_aspect()`. Default is `equal`. See full
        description at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_aspect.html
    **kwargs : optional args
        Alternate default values. Values that can be overwritten are the following:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * min_dyad_lw
        * max_dyad_lw
        * node_fc_cmap
        * node_ec_cmap
        * dyad_color_cmap
        * edge_fc_cmap

    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph()
    >>> H.add_edges_from([[1,2,3],[3,4],[4,5,6,7],[7,8,9,10,11]])
    >>> ax = xgi.draw(H, pos=xgi.barycenter_spring_layout(H))

    See Also
    --------
    draw_nodes
    draw_hyperedges
    draw_simplices
    draw_node_labels
    draw_hyperedge_labels

    """

    settings = {
        "min_node_size": 5,
        "max_node_size": 30,
        "min_dyad_lw": 2.0,
        "max_dyad_lw": 10.0,
        "min_node_lw": 0,
        "max_node_lw": 5,
        "node_fc_cmap": cm.Reds,
        "node_ec_cmap": cm.Greys,
        "edge_fc_cmap": cm.Blues,
        "dyad_color_cmap": cm.Greys,
    }

    settings.update(kwargs)

    if edge_fc is None:
        edge_fc = H.edges.size

    ax, pos = _draw_init(H, ax, pos)

    if not max_order:
        max_order = max_edge_order(H)

    if isinstance(H, SimplicialComplex):
        draw_simplices(
            H,
            pos,
            ax,
            dyad_color,
            dyad_lw,
            edge_fc,
            max_order,
            settings,
            hyperedge_labels,
            **kwargs,
        )
    elif isinstance(H, Hypergraph):
        draw_hyperedges(
            H,
            pos,
            ax,
            dyad_color,
            dyad_lw,
            edge_fc,
            max_order,
            settings,
            hyperedge_labels,
            **kwargs,
        )
    else:
        raise XGIError("The input must be a SimplicialComplex or Hypergraph")

    ax, node_collection = draw_nodes(
        H=H,
        pos=pos,
        ax=ax,
        node_fc=node_fc,
        node_ec=node_ec,
        node_lw=node_lw,
        node_size=node_size,
        node_shape=node_shape,
        zorder=max_order,
        params=settings,
        node_labels=node_labels,
        **kwargs,
    )

    # compute axis limits
    _update_lims(pos, ax)

    ax.set_aspect(aspect, "datalim")

    return ax, node_collection


def draw_nodes(
    H,
    pos=None,
    ax=None,
    node_fc="white",
    node_ec="black",
    node_lw=1,
    node_size=15,
    node_shape="o",
    node_fc_cmap="Reds",
    vmin=None,
    vmax=None,
    zorder=None,
    params=dict(),
    node_labels=False,
    rescale_sizes=True,
    **kwargs,
):
    """Draw the nodes of a hypergraph

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        Higher-order network to plot.
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    pos : dict, optional
        If passed, this dictionary of positions node_id:(x,y) is used for placing the
        0-simplices.  If None (default), use the `barycenter_spring_layout` to compute
        the positions.
    node_fc : str, iterable, or NodeStat, optional
        Color of the nodes.  If str, use the same color for all nodes. If other iterable,
        or NodeStat, assume the colors are specified in the same order as the nodes are
        found in H.nodes. By default, "white".
    node_ec : color or sequence of colors, optional
        Color of node borders. If color, use the same color for all nodes. If sequence
        of colors, assume the colors are specified in the same order as the nodes are
        found in H.nodes. By default, "black".
    node_lw : int, float, iterable, or NodeStat, optional
        Line width of the node borders in pixels.  If int or float, use the same width
        for all node borders.  If iterable or NodeStat, assume the widths are specified
        in the same order as the nodes are found in H.nodes. Values are clipped below
        and above by min_node_lw and max_node_lw, respectively. By default, 1.
    node_size : int, float, iterable, or NodeStat, optional
        Radius of the nodes in pixels.  If int or float, use the same radius for all
        nodes. If iterable or NodeStat, assume the radiuses are specified in the same
        order as the nodes are found in H.nodes. Values are clipped below
        and above by min_node_size and max_node_size, respectively. By default, 15.
    node_shape :  string, optional
        The shape of the node. Specification is as matplotlib.scatter
        marker. Default is "o".
    node_fc_cmap : colormap
        Colormap for mapping node colors. By default, "Reds". Ignored, if `node_fc` is
        a str (single color).
    vmin : float or None
        Minimum for the node_fc_cmap scaling. By default, None.
    vmax : float or None
        Maximum for the node_fc_cmap scaling. By default, None.
    zorder : int
        The layer on which to draw the nodes.
    node_labels : bool or dict
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
    rescale_sizes: bool, optional
        If True, linearly interpolate `node_size` and `node_lw` between min/max values
        (5/30 for size, 0/5 for lw) that can be changed in the other argument `params`.
        If `node_size` (`node_lw`) is a single value, `interpolate_sizes` is ignored
        for it. By default, True.
    params : dict
        Default parameters used if `interpolate_sizes` is True.
        Keys to override default settings:
        * "min_node_size" (default: 5)
        * "max_node_size" (default: 30)
        * "min_node_lw" (default: 0)
        * "max_node_lw" (default: 5)
    kwargs : optional keywords
        See `draw_node_labels` for a description of optional keywords.

    Returns
    -------
    ax : matplotlib Axes
        Axes plotted on
    node_collection : matplotlib PathCollection
        Collection containing the nodes

    See Also
    --------
    draw
    draw_hyperedges
    draw_simplices
    draw_node_labels
    draw_hyperedge_labels

    Notes
    -----

    * If nodes are colored with a cmap, the `node_collection` returned
    can be used to easily plot a colorbar corresponding to the node
    colors. Simply do `plt.colorbar(node_collection)`.

    * Nodes with nonfinite `node_fc` (i.e. `inf`, `-inf` or `nan` are drawn
    with the bad colormap color (see `plotnonfinitebool` in `plt.scatter` and
    Colormap.set_bad from Matplotlib).

    """

    settings = {
        "min_node_size": 5,
        "max_node_size": 30,
        "min_node_lw": 0,
        "max_node_lw": 5,
    }

    settings.update(params)
    settings.update(kwargs)

    # avoid matplotlib scatter UserWarning "Parameters 'cmap' will be ignored"
    if isinstance(node_fc, str):
        node_fc_cmap = None

    ax, pos = _draw_init(H, ax, pos)

    # convert pos to format convenient for scatter
    try:
        xy = np.asarray([pos[v] for v in H.nodes])
    except KeyError as err:
        raise XGIError(f"Node {err} has no position.") from err

    # convert all formats to ndarray
    node_size = _draw_arg_to_arr(node_size)
    node_fc = _draw_arg_to_arr(node_fc)
    node_lw = _draw_arg_to_arr(node_lw)

    # check validity of input values
    if np.any(node_size < 0):
        raise ValueError("node_size cannot contain negative values.")
    if np.any(node_lw < 0):
        raise ValueError("node_lw cannot contain negative values.")

    # interpolate if needed
    if rescale_sizes and isinstance(node_size, np.ndarray):
        node_size = _interp_draw_arg(
            node_size, settings["min_node_size"], settings["max_node_size"]
        )
    if rescale_sizes and isinstance(node_lw, np.ndarray):
        node_lw = _interp_draw_arg(
            node_lw, settings["min_node_lw"], settings["max_node_lw"]
        )

    node_size = node_size**2

    # plot
    node_collection = ax.scatter(
        x=xy[:, 0],
        y=xy[:, 1],
        s=node_size,
        marker=node_shape,
        c=node_fc,
        cmap=node_fc_cmap,
        vmin=vmin,
        vmax=vmax,
        edgecolors=node_ec,
        linewidths=node_lw,
        zorder=zorder,
        plotnonfinite=True,  # plot points with nonfinite color
    )

    if node_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_node_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "node_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_node_labels(H, pos, node_labels, ax_nodes=ax, **label_kwds)

    # compute axis limits
    _update_lims(pos, ax)

    return ax, node_collection


def draw_hyperedges(
    H,
    pos=None,
    ax=None,
    dyad_color="black",
    dyad_lw=1.5,
    edge_fc=None,
    max_order=None,
    settings=None,
    hyperedge_labels=False,
    **kwargs,
):
    """Draw hyperedges.

    Parameters
    ----------
    H : Hypergraph
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    pos : dict, optional
        If passed, this dictionary of positions node_id:(x,y) is used for placing the
        0-simplices.  If None (default), use the `barycenter_spring_layout` to compute
        the positions.
    dyad_color : str, dict, iterable, or EdgeStat, optional
        Color of the dyadic links.  If str, use the same color for all edges. If a dict,
        must contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use
        a colormap (specified with dyad_color_cmap) associated to it. By default,
        "black".
    dyad_lw : int, float, dict, iterable, or EdgeStat, optional
        Line width of edges of order 1 (dyadic links).  If int or float, use the same
        width for all edges.  If a dict, must contain (edge_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the edges are
        found in H.edges. If EdgeStat, use a monotonic linear interpolation defined
        between min_dyad_lw and max_dyad_lw. By default, 1.5.
    edge_fc : str, dict, iterable, or EdgeStat, optional
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict,
        must contain (edge_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the hyperedges are found in H.edges. If
        EdgeStat, use the colormap specified with edge_fc_cmap. If None (default), color
        by edge size.
    max_order : int, optional
        Maximum of hyperedges to plot. By default, None.
    hyperedge_labels : bool or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  By default, None.
    settings : dict
        Default parameters. Keys that may be useful to override default settings:
        * min_dyad_lw
        * max_dyad_lw
        * dyad_color_cmap
        * edge_fc_cmap
    kwargs : optional keywords
        See `draw_hyperedge_labels` for a description of optional keywords.

    Raises
    ------
    XGIError
        If a SimplicialComplex is passed.

    See Also
    --------
    draw
    draw_nodes
    draw_simplices
    draw_node_labels
    draw_hyperedge_labels

    """

    ax, pos = _draw_init(H, ax, pos)

    if max_order is None:
        max_order = max_edge_order(H)

    if edge_fc is None:
        edge_fc = H.edges.size

    if settings is None:
        settings = {
            "min_dyad_lw": 2.0,
            "max_dyad_lw": 10.0,
            "edge_fc_cmap": cm.Blues,
            "dyad_color_cmap": cm.Greys,
        }

    settings.update(kwargs)

    dyad_color = _color_arg_to_dict(dyad_color, H.edges, settings["dyad_color_cmap"])
    dyad_lw = _scalar_arg_to_dict(
        dyad_lw, H.edges, settings["min_dyad_lw"], settings["max_dyad_lw"]
    )

    edge_fc = _color_arg_to_dict(edge_fc, H.edges, settings["edge_fc_cmap"])

    # Looping over the hyperedges of different order (reversed) -- nodes will be plotted
    # separately
    for id, he in H.edges.members(dtype=dict).items():
        d = len(he) - 1
        if d > max_order:
            continue
        if d == 1:
            # Drawing the edges
            he = list(he)
            x_coords = [pos[he[0]][0], pos[he[1]][0]]
            y_coords = [pos[he[0]][1], pos[he[1]][1]]
            line = plt.Line2D(
                x_coords,
                y_coords,
                color=dyad_color[id],
                lw=dyad_lw[id],
                zorder=max_order - 1,
            )
            ax.add_line(line)
        else:
            # Hyperedges of order d (d=2: triangles, etc.)
            # Filling the polygon
            coordinates = [[pos[n][0], pos[n][1]] for n in he]
            # Sorting the points counterclockwise (needed to have the correct filling)
            sorted_coordinates = _CCW_sort(coordinates)
            obj = plt.Polygon(
                sorted_coordinates,
                facecolor=edge_fc[id],
                alpha=0.4,
                zorder=max_order - d,
            )
            ax.add_patch(obj)

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_hyperedge_labels(H, pos, hyperedge_labels, ax_edges=ax, **label_kwds)

    # compute axis limits
    _update_lims(pos, ax)

    return ax


def draw_simplices(
    SC,
    pos=None,
    ax=None,
    dyad_color="black",
    dyad_lw=1.5,
    edge_fc=None,
    max_order=None,
    settings=None,
    hyperedge_labels=False,
    **kwargs,
):
    """Draw maximal simplices and pairwise faces.

    Parameters
    ----------
    SC : SimplicialComplex
        Simplicial complex to draw
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    pos : dict, optional
        If passed, this dictionary of positions node_id:(x,y) is used for placing the
        0-simplices.  If None (default), use the `barycenter_spring_layout` to compute
        the positions.
    dyad_color : str, dict, iterable, or EdgeStat, optional
        Color of the dyadic links.  If str, use the same color for all edges. If a dict,
        must contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use
        a colormap (specified with dyad_color_cmap) associated to it. By default,
        "black".
    dyad_lw : int, float, dict, iterable, or EdgeStat, optional
        Line width of edges of order 1 (dyadic links).  If int or float, use the same
        width for all edges.  If a dict, must contain (edge_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the edges are
        found in H.edges. If EdgeStat, use a monotonic linear interpolation defined
        between min_dyad_lw and max_dyad_lw. By default, 1.5.
    edge_fc : str, dict, iterable, or EdgeStat, optional
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict,
        must contain (edge_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the hyperedges are found in H.edges. If
        EdgeStat, use the colormap specified with edge_fc_cmap. If None (default), color
        by simplex size.
    max_order : int, optional
        Maximum of hyperedges to plot. By default, None.
    hyperedge_labels : bool or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  Note, we plot only the maximal simplices so if you pass a dict be
        careful to match its keys with the new edge ids in the converted
        SimplicialComplex. These may differ from the edge ids in the given SC. By
        default, False.
    settings : dict
        Default parameters. Keys that may be useful to override default settings:
        * min_dyad_lw
        * max_dyad_lw
        * dyad_color_cmap
        * edge_fc_cmap
    kwargs : optional keywords
        See `draw_hyperedge_labels` for a description of optional keywords.


    Raises
    ------
    XGIError
        If a Hypergraph is passed.

    See Also
    --------
    draw
    draw_nodes
    draw_hyperedges
    draw_node_labels
    draw_hyperedge_labels

    """

    if max_order:
        max_edges = SC.edges.filterby("order", max_order, "leq").members()
        SC = SimplicialComplex(max_edges)  # SC without simplices larger than max_order

    # Plot only the maximal simplices, thus let's convert the SC to H
    H_ = convert.from_max_simplices(SC)

    if not max_order:
        max_order = max_edge_order(H_)

    ax, pos = _draw_init(H_, ax, pos)

    if edge_fc is None:
        edge_fc = H_.edges.size

    if settings is None:
        settings = {
            "min_dyad_lw": 2.0,
            "max_dyad_lw": 10.0,
            "edge_fc_cmap": cm.Blues,
            "dyad_color_cmap": cm.Greys,
        }

    settings.update(kwargs)

    dyad_color = _color_arg_to_dict(dyad_color, H_.edges, settings["dyad_color_cmap"])
    dyad_lw = _scalar_arg_to_dict(
        dyad_lw,
        H_.edges,
        settings["min_dyad_lw"],
        settings["max_dyad_lw"],
    )

    edge_fc = _color_arg_to_dict(edge_fc, H_.edges, settings["edge_fc_cmap"])

    # Looping over the hyperedges of different order (reversed) -- nodes will be plotted
    # separately
    for id, he in H_.edges.members(dtype=dict).items():
        d = len(he) - 1

        if d == 1:
            # Drawing the edges
            he = list(he)
            x_coords = [pos[he[0]][0], pos[he[1]][0]]
            y_coords = [pos[he[0]][1], pos[he[1]][1]]

            line = plt.Line2D(
                x_coords,
                y_coords,
                color=dyad_color[id],
                lw=dyad_lw[id],
                zorder=max_order - 1,
            )
            ax.add_line(line)
        else:
            # Hyperedges of order d (d=2: triangles, etc.)
            # Filling the polygon
            coordinates = [[pos[n][0], pos[n][1]] for n in he]
            # Sorting the points counterclockwise (needed to have the correct filling)
            sorted_coordinates = _CCW_sort(coordinates)
            obj = plt.Polygon(
                sorted_coordinates,
                facecolor=edge_fc[id],
                alpha=0.4,
                zorder=max_order - d,
            )
            ax.add_patch(obj)
            # Drawing all the edges within
            for i, j in combinations(sorted_coordinates, 2):
                x_coords = [i[0], j[0]]
                y_coords = [i[1], j[1]]
                line = plt.Line2D(
                    x_coords,
                    y_coords,
                    color=dyad_color[id],
                    lw=dyad_lw[id],
                    zorder=max_order - 1,
                )
                ax.add_line(line)

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_hyperedge_labels(H_, pos, hyperedge_labels, ax_edges=ax, **label_kwds)

    # compute axis limits
    _update_lims(pos, ax)

    return ax


def draw_node_labels(
    H,
    pos,
    node_labels=False,
    font_size_nodes=10,
    font_color_nodes="black",
    font_family_nodes="sans-serif",
    font_weight_nodes="normal",
    alpha_nodes=None,
    bbox_nodes=None,
    horizontalalignment_nodes="center",
    verticalalignment_nodes="center",
    ax_nodes=None,
    clip_on_nodes=True,
):
    """Draw node labels on the hypergraph or simplicial complex.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex.
    pos : dict
        Dictionary of positions node_id:(x,y).
    node_labels : bool or dict, optional
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
        By default, False.
    font_size_nodes : int, optional
        Font size for text labels, by default 10.
    font_color_nodes : str, optional
        Font color string, by default "black".
    font_family_nodes : str, optional
        Font family, by default "sans-serif".
    font_weight_nodes : str (default='normal')
        Font weight.
    alpha_nodes : float, optional
        The text transparency, by default None.
    bbox_nodes : Matplotlib bbox, optional
        Specify text box properties (e.g. shape, color etc.) for node labels.
        When it is None (default), use Matplotlib's ax.text default
    horizontalalignment_nodes : str, optional
        Horizontal alignment {'center', 'right', 'left'}.
        By default, "center".
    verticalalignment_nodes : str, optional
        Vertical alignment {'center', 'top', 'bottom', 'baseline', 'center_baseline'}.
        By default, "center".
    ax_nodes : matplotlib.pyplot.axes, optional
        Draw the graph in the specified Matplotlib axes.
        By default, None.
    clip_on_nodes : bool, optional
        Turn on clipping of node labels at axis boundaries.
        By default, True.

    Returns
    -------
    dict
        `dict` of labels keyed by node id.

    See Also
    --------
    draw
    draw_nodes
    draw_hyperedges
    draw_simplices
    draw_hyperedge_labels
    """
    if ax_nodes is None:
        ax = plt.gca()
    else:
        ax = ax_nodes

    if node_labels is True:
        node_labels = {id: id for id in H.nodes}

    # Plot the labels in the last layer
    zorder = max_edge_order(H) + 1

    text_items = {}
    for idx, label in node_labels.items():
        (x, y) = pos[idx]

        if not isinstance(label, str):
            label = str(label)

        t = ax.text(
            x,
            y,
            label,
            size=font_size_nodes,
            color=font_color_nodes,
            family=font_family_nodes,
            weight=font_weight_nodes,
            alpha=alpha_nodes,
            horizontalalignment=horizontalalignment_nodes,
            verticalalignment=verticalalignment_nodes,
            transform=ax.transData,
            bbox=bbox_nodes,
            clip_on=clip_on_nodes,
            zorder=zorder,
        )
        text_items[idx] = t

    return text_items


def draw_hyperedge_labels(
    H,
    pos,
    hyperedge_labels=False,
    font_size_edges=10,
    font_color_edges="black",
    font_family_edges="sans-serif",
    font_weight_edges="normal",
    alpha_edges=None,
    bbox_edges=None,
    horizontalalignment_edges="center",
    verticalalignment_edges="center",
    ax_edges=None,
    rotate_edges=False,
    clip_on_edges=True,
):
    """Draw hyperedge labels on the hypegraph or simplicial complex.

    Parameters
    ----------
    H : Hypergraph.
    pos : dict
        Dictionary of positions node_id:(x,y).
    hyperedge_labels : bool or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  By default, False.
    font_size_edges : int, optional
        Font size for text labels, by default 10.
    font_color_edges : str, optional
        Font color string, by default "black".
    font_family_edges : str (default='sans-serif')
        Font family.
    font_weight_edges : str (default='normal')
        Font weight.
    alpha_edges : float, optional
        The text transparency, by default None.
    bbox_edges : Matplotlib bbox, optional
        Specify text box properties (e.g. shape, color etc.) for edge labels.
        By default, {boxstyle='round', ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0)}
    horizontalalignment_edges : str, optional
        Horizontal alignment {'center', 'right', 'left'}.
        By default, "center".
    verticalalignment_edges: str, optional
        Vertical alignment {'center', 'top', 'bottom', 'baseline', 'center_baseline'}.
        By default, "center".
    ax_edges : matplotlib.pyplot.axes, optional
        Draw the graph in the specified Matplotlib axes. By default, None.
    rotate_edges : bool, optional
        Rotate edge labels for dyadic links to lie parallel to edges, by default False.
    clip_on_edges: bool, optional
        Turn on clipping of hyperedge labels at axis boundaries, by default True.

    Returns
    -------
    dict
        `dict` of labels keyed by hyperedge id.

    See Also
    --------
    draw
    draw_nodes
    draw_hyperedges
    draw_simplices
    draw_node_labels

    """
    if ax_edges is None:
        ax = plt.gca()
    else:
        ax = ax_edges

    if hyperedge_labels is True:
        hyperedge_labels = {id: id for id in H.edges}

    text_items = {}
    for id, label in hyperedge_labels.items():
        he = H.edges.members(id)
        coordinates = [[pos[n][0], pos[n][1]] for n in he]
        x, y = np.mean(coordinates, axis=0)

        if len(he) == 2:
            # Rotate edge labels for dyadic links to lie parallel to edges
            if rotate_edges:
                x_diff, y_diff = np.subtract(coordinates[1], coordinates[0])
                angle = np.arctan2(y_diff, x_diff) / (2.0 * np.pi) * 360
                # Make label orientation "right-side-up"
                if angle > 90:
                    angle -= 180
                if angle < -90:
                    angle += 180
                # Transform data coordinate angle to screen coordinate angle
                xy = np.array((x, y))
                trans_angle = ax.transData.transform_angles(
                    np.array((angle,)), xy.reshape((1, 2))
                )[0]
            else:
                trans_angle = 0.0
        else:
            trans_angle = 0.0

        # Use default box of white with white border
        if bbox_edges is None:
            bbox = dict(boxstyle="round", ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0))
        else:
            bbox = bbox_edges

        if not isinstance(label, str):
            label = str(label)

        t = ax.text(
            x,
            y,
            label,
            size=font_size_edges,
            color=font_color_edges,
            family=font_family_edges,
            weight=font_weight_edges,
            alpha=alpha_edges,
            horizontalalignment=horizontalalignment_edges,
            verticalalignment=verticalalignment_edges,
            rotation=trans_angle,
            transform=ax.transData,
            bbox=bbox,
            clip_on=clip_on_edges,
        )
        text_items[id] = t

    return text_items


def _draw_hull(node_pos, ax, edges_ec, facecolor, alpha, zorder, radius):
    """Draw a convex hull encompassing the nodes in node_pos

    Parameters
    ----------
    node_pos : np.array
        nx2 dimensional array containing positions of the nodes
    ax : matplotlib.pyplot.axes
    edges_ec : str
        Color of the border of the convex hull
    facecolor : str
        Filling color of the convex hull
    alpha : float
        Transparency of the convex hull
    radius : float
        Radius of the convex hull in the vicinity of the nodes.

    Returns
    -------
    ax : matplotlib.pyplot.axes

    """

    thetas = np.linspace(0, 2 * np.pi, num=100, endpoint=False)
    offsets = radius * np.array([np.cos(thetas), np.sin(thetas)]).T
    points = np.vstack([p + offsets for p in node_pos])
    points = np.vstack([node_pos, points])

    hull = ConvexHull(points)

    for simplex in hull.simplices:
        ax.plot(points[simplex, 0], points[simplex, 1], color=edges_ec, zorder=zorder)
    ax.fill(
        points[hull.vertices, 0],
        points[hull.vertices, 1],
        color=facecolor,
        alpha=alpha,
        zorder=zorder,
    )

    return ax


def draw_hypergraph_hull(
    H,
    pos=None,
    ax=None,
    dyad_color="black",
    edge_fc=None,
    edge_ec=None,
    node_fc="tab:blue",
    node_ec="black",
    node_lw=1,
    node_size=7,
    node_shape="o",
    max_order=None,
    node_labels=False,
    hyperedge_labels=False,
    radius=0.05,
    aspect="equal",
    **kwargs,
):
    """Draw hypergraphs displaying the hyperedges of order k>1 as convex hulls


    Parameters
    ----------
    H : Hypergraph
    pos : dict, optional
        If passed, this dictionary of positions node_id:(x,y) is used for placing the
        nodes.  If None (default), use the `barycenter_spring_layout` to compute the
        positions.
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    dyad_color : str, dict, iterable, or EdgeStat, optional
        Color of the dyadic links.  If str, use the same color for all edges. If a dict,
        must contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use
        a colormap (specified with dyad_color_cmap) associated to it. By default,
        "black".
    edge_fc : str, dict, iterable, or EdgeStat, optional
        Color of the hyperedges of order k>1.  If str, use the same color for all
        hyperedges of order k>1.  If a dict, must contain (edge_id: color_str) pairs.
        If other iterable, assume the colors are specified in the same order as the
        hyperedges are found in H.edges. If EdgeStat, use the colormap specified with
        edge_fc_cmap. If None (default), use the H.edges.size.
    edge_ec : str, dict, iterable, or EdgeStat, optional
        Color of the borders of the hyperdges of order k>1.  If str, use the same color
        for all edges. If a dict, must contain (edge_id: color_str) pairs.  If iterable,
        assume the colors are specified in the same order as the edges are found in
        H.edges. If EdgeStat, use a colormap (specified with edge_ec_cmap) associated to
        it. If None (default), use the H.edges.size.
    node_fc : node_fc : str, dict, iterable, or NodeStat, optional
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use
        the colormap specified with node_fc_cmap. By default, "tab:blue".
    node_ec : str, dict, iterable, or NodeStat, optional
        Color of node borders.  If str, use the same color for all nodes.  If a dict,
        must contain (node_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap. By default, "black".
    node_lw : int, float, dict, iterable, or EdgeStat, optional
        Line width of the node borders in pixels.  If int or float, use the same width
        for all node borders.  If a dict, must contain (node_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the nodes are
        found in H.nodes. If NodeStat, use a monotonic linear interpolation defined
        between min_node_lw and max_node_lw. By default, 1.
    node_size : int, float, dict, iterable, or NodeStat, optional
        Radius of the nodes in pixels.  If int or float, use the same radius for all
        nodes.  If a dict, must contain (node_id: radius) pairs.  If iterable, assume
        the radiuses are specified in the same order as the nodes are found in
        H.nodes. If NodeStat, use a monotonic linear interpolation defined between
        min_node_size and max_node_size. By default, 7.
    node_shape :  string, optional
        The shape of the node. Specification is as matplotlib.scatter
        marker. Default is "o".
    max_order : int, optional
        Maximum of hyperedges to plot. If None (default), plots all orders.
    node_labels : bool, or dict, optional
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
        By default, False
    hyperedge_labels : bool, or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  By default, False.
    radius : float, optional
        Radius of the convex hull in the vicinity of the nodes, by default 0.05.
    aspect : {"auto", "equal"} or float, optional
        Set the aspect ratio of the axes scaling, i.e. y/x-scale. `aspect` is passed
        directly to matplotlib's `ax.set_aspect()`. Default is `equal`. See full
        description at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_aspect.html
    **kwargs : optional args
        Alternate default values. Values that can be overwritten are the following:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * node_fc_cmap
        * node_ec_cmap
        * dyad_color_cmap
        * edge_fc_cmap
        * edge_ec_cmap
        * alpha

    Returns
    -------
    ax : matplotlib.pyplot.axes

    See Also
    --------
    draw

    """

    settings = {
        "min_node_size": 5.0,
        "max_node_size": 30.0,
        "min_node_lw": 1.0,
        "max_node_lw": 5.0,
        "node_fc_cmap": cm.Reds,
        "node_ec_cmap": cm.Greys,
        "dyad_color_cmap": cm.Greys,
        "edge_fc_cmap": cm.Blues,
        "edge_ec_cmap": cm.Greys,
        "alpha": 0.4,
    }

    alpha = settings["alpha"]

    if edge_fc is None:
        edge_fc = H.edges.size

    edge_fc = _color_arg_to_dict(edge_fc, H.edges, settings["edge_fc_cmap"])

    if edge_ec is None:
        edge_ec = H.edges.size

    edge_ec = _color_arg_to_dict(edge_ec, H.edges, settings["edge_ec_cmap"])

    settings.update(kwargs)

    ax, pos = _draw_init(H, ax, pos)

    if not max_order:
        max_order = max_edge_order(H)

    dyad_color = _color_arg_to_dict(dyad_color, H.edges, settings["dyad_color_cmap"])

    for id, he in H._edge.items():
        d = len(he) - 1
        if d > max_order:
            continue
        if d == 1:
            # Drawing the edges
            he = list(he)
            x_coords = [pos[he[0]][0], pos[he[1]][0]]
            y_coords = [pos[he[0]][1], pos[he[1]][1]]
            line = plt.Line2D(
                x_coords,
                y_coords,
                color=dyad_color[id],
                zorder=max_order - 1,
                alpha=1,
            )
            ax.add_line(line)

        else:
            coordinates = [[pos[n][0], pos[n][1]] for n in he]
            _draw_hull(
                node_pos=np.array(coordinates),
                ax=ax,
                edges_ec=edge_ec[id],
                facecolor=edge_fc[id],
                alpha=alpha,
                zorder=max_order - d,
                radius=radius,
            )

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        draw_hyperedge_labels(H, pos, hyperedge_labels, ax_edges=ax, **label_kwds)

    ax, node_collection = draw_nodes(
        H=H,
        pos=pos,
        ax=ax,
        node_fc=node_fc,
        node_ec=node_ec,
        node_lw=node_lw,
        node_size=node_size,
        node_shape=node_shape,
        zorder=max_order,
        params=settings,
        node_labels=node_labels,
        **kwargs,
    )

    # compute axis limits
    _update_lims(pos, ax)

    ax.set_aspect(aspect, "datalim")

    return ax, node_collection


def draw_multilayer(
    H,
    pos=None,
    ax=None,
    dyad_color="black",
    dyad_lw=0.5,
    edge_fc=None,
    node_fc="white",
    node_ec="black",
    node_lw=0.5,
    node_size=5,
    plane_color="grey",
    max_order=None,
    conn_lines=True,
    conn_lines_style="dotted",
    width=5,
    height=5,
    h_angle=10,
    v_angle=20,
    sep=1,
    **kwargs,
):
    """Draw a hypergraph or simplicial complex visualized in 3D
    showing hyperedges/simplices of different orders on superimposed layers.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex.
        Higher-order network to plot.
    pos : dict or None, optional
        The positions of the nodes in the multilayer network.
        If None, a default layout will be computed using
        xgi.barycenter_spring_layout(). Default is None.
    ax : matplotlib Axes3DSubplot or None, optional
        The subplot to draw the visualization on.
        If None, a new subplot will be created. Default is None.
    dyad_color : str, dict, iterable, or EdgeStat, optional
        Color of the dyadic links.  If str, use the same color for all edges. If a dict,
        must contain (edge_id: color_str) pairs.  If iterable, assume the colors are
        specified in the same order as the edges are found in H.edges. If EdgeStat, use
        a colormap (specified with dyad_color_cmap) associated to it. By default,
        "black".
    dyad_lw : int, float, dict, iterable, or EdgeStat, optional
        Line width of edges of order 1 (dyadic links).  If int or float, use the same
        width for all edges.  If a dict, must contain (edge_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the edges are
        found in H.edges. If EdgeStat, use a monotonic linear interpolation defined
        between min_dyad_lw and max_dyad_lw. By default, 0.5.
    edge_fc : str, dict, iterable, or EdgeStat, optional
        Color of the hyperedges.  If str, use the same color for all nodes.  If a dict,
        must contain (edge_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the hyperedges are found in H.edges. If
        EdgeStat, use the colormap specified with edge_fc_cmap. If None (default), use
        the H.edges.size.
    node_fc : str, dict, iterable, or NodeStat, optional
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use
        the colormap specified with node_fc_cmap. By default, "white".
    node_ec : str, dict, iterable, or NodeStat, optional
        Color of node borders.  If str, use the same color for all nodes.  If a dict,
        must contain (node_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap. By default, "black".
    node_lw : int, float, dict, iterable, or NodeStat, optional
        Line width of the node borders in pixels.  If int or float, use the same width
        for all node borders.  If a dict, must contain (node_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the nodes are
        found in H.nodes. If NodeStat, use a monotonic linear interpolation defined
        between min_node_lw and max_node_lw. By default, 1.
    node_size : int, float, dict, iterable, or NodeStat, optional
        Radius of the nodes in pixels.  If int or float, use the same radius for all
        nodes.  If a dict, must contain (node_id: radius) pairs.  If iterable, assume
        the radiuses are specified in the same order as the nodes are found in
        H.nodes. If NodeStat, use a monotonic linear interpolation defined between
        min_node_size and max_node_size. By default, 5.
    plane_color : color (str or tuple) or iterable (dict, list, or numpy array), optional
        Color of each plane. If a dict, must contain (edge size: color) pairs.
        By default, "grey".
    max_order : int, optional
        Maximum of hyperedges to plot. If None (default), plots all orders.
    conn_lines : bool, optional
        Whether to draw connections between layers. Default is True.
    conn_lines_style : str, optional
        The linestyle of the connections between layers. Default is 'dotted'.
    width : float, optional
        The width of the figure in inches. Default is 5.
    height : float, optional
        The height of the figure in inches. Default is 5.
    h_angle : float, optional
        The rotation angle around the horizontal axis in degrees. Default is 10.
    v_angle : float, optional
        The rotation angle around the vertical axis in degrees. Default is 0.
    sep : float, optional
        The separation between layers. Default is 1.
    **kwargs : optional args
        Alternate default values. Values that can be overwritten are the following:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * min_dyad_lw
        * max_dyad_lw
        * node_fc_cmap
        * node_ec_cmap
        * dyad_color_cmap
        * edge_fc_cmap

    Returns
    -------
    ax : matplotlib Axes3DSubplot
        The subplot with the multilayer network visualization.
    """
    settings = {
        "min_node_size": 10.0,
        "max_node_size": 30.0,
        "min_dyad_lw": 2.0,
        "max_dyad_lw": 10.0,
        "min_node_lw": 1.0,
        "max_node_lw": 5.0,
        "node_fc_cmap": cm.Reds,
        "node_ec_cmap": cm.Greys,
        "edge_fc_cmap": cm.Blues,
        "dyad_color_cmap": cm.Greys,
        "plane_color_cmap": cm.Greys,
    }

    settings.update(kwargs)

    if edge_fc is None:
        edge_fc = H.edges.size

    if pos is None:
        pos = barycenter_spring_layout(H)

    if ax is None:
        _, ax = plt.subplots(
            1, 1, figsize=(width, height), dpi=600, subplot_kw={"projection": "3d"}
        )

    s = unique_edge_sizes(H)
    if max_order is None:
        max_order = max(s) - 1
    else:
        max_order = min(max_order, max(s) - 1)
    min_order = min(s) - 1

    xs, ys = zip(*pos.values())

    dyad_color = _color_arg_to_dict(dyad_color, H.edges, settings["dyad_color_cmap"])
    dyad_lw = _scalar_arg_to_dict(
        dyad_lw, H.edges, settings["min_dyad_lw"], settings["max_dyad_lw"]
    )

    edge_fc = _color_arg_to_dict(edge_fc, H.edges, settings["edge_fc_cmap"])

    node_fc = _color_arg_to_dict(node_fc, H.nodes, settings["node_fc_cmap"])
    node_ec = _color_arg_to_dict(node_ec, H.nodes, settings["node_ec_cmap"])
    node_lw = _scalar_arg_to_dict(
        node_lw,
        H.nodes,
        settings["min_node_lw"],
        settings["max_node_lw"],
    )
    node_size = _scalar_arg_to_dict(
        node_size, H.nodes, settings["min_node_size"], settings["max_node_size"]
    )

    plane_color = _color_arg_to_dict(
        plane_color,
        [i for i in range(min_order, max_order + 1)],
        settings["plane_color_cmap"],
    )

    for id, he in H.edges.members(dtype=dict).items():
        d = len(he) - 1
        zs = d * sep

        # dyads
        if d > max_order:
            continue

        if d == 1:
            he = list(he)
            x1 = [pos[he[0]][0], pos[he[0]][1], zs]
            x2 = [pos[he[1]][0], pos[he[1]][1], zs]
            l = Line3DCollection(
                [(x1, x2)],
                color=dyad_color[id],
                linewidth=dyad_lw[id],
            )
            ax.add_collection3d(l)
        # higher-orders
        else:
            poly = []
            vertices = np.array([[pos[i][0], pos[i][1], zs] for i in he])
            vertices = _CCW_sort(vertices)
            poly.append(vertices)
            poly = Poly3DCollection(
                poly,
                zorder=d - 1,
                color=edge_fc[id],
                alpha=0.5,
                edgecolor=None,
            )
            ax.add_collection3d(poly)

    # now draw by order
    # draw lines connecting points on the different planes
    if conn_lines:
        lines3d_between = [
            (list(pos[i]) + [min_order * sep], list(pos[i]) + [max_order * sep])
            for i in H.nodes
        ]
        between_lines = Line3DCollection(
            lines3d_between,
            zorder=d,
            color=".5",
            alpha=0.4,
            linestyle=conn_lines_style,
            linewidth=1,
        )
        ax.add_collection3d(between_lines)

    (x, y, s, c, ec, lw,) = zip(
        *[
            (
                pos[i][0],
                pos[i][1],
                node_size[i] ** 2,
                node_fc[i],
                node_ec[i],
                node_lw[i],
            )
            for i in H.nodes
        ]
    )
    for d in range(min_order, max_order + 1):
        # draw nodes
        z = [sep * d] * H.num_nodes
        ax.scatter(
            x,
            y,
            z,
            s=s,
            c=c,
            edgecolors=ec,
            linewidths=lw,
            zorder=max_order + 1,
            alpha=1,
        )

        # draw surfaces corresponding to the different orders
        xdiff = np.max(xs) - np.min(xs)
        ydiff = np.max(ys) - np.min(ys)
        ymin = np.min(ys) - ydiff * 0.1
        ymax = np.max(ys) + ydiff * 0.1
        xmin = np.min(xs) - xdiff * 0.1 * (width / height)
        xmax = np.max(xs) + xdiff * 0.1 * (width / height)
        xx, yy = np.meshgrid([xmin, xmax], [ymin, ymax])
        zz = np.zeros(xx.shape) + d * sep
        ax.plot_surface(
            xx,
            yy,
            zz,
            color=plane_color[d],
            alpha=0.1,
            zorder=d,
        )

    ax.view_init(h_angle, v_angle)
    ax.set_ylim(np.min(ys) - ydiff * 0.1, np.max(ys) + ydiff * 0.1)
    ax.set_xlim(np.min(xs) - xdiff * 0.1, np.max(xs) + xdiff * 0.1)
    ax.set_axis_off()

    return ax


def draw_dihypergraph(
    DH,
    ax=None,
    lines_fc=None,
    lines_lw=1.5,
    line_head_width=0.05,
    node_fc="white",
    node_ec="black",
    node_lw=1,
    node_size=15,
    edge_marker_toggle=True,
    edge_marker_fc=None,
    edge_marker_ec=None,
    edge_marker="s",
    edge_marker_lw=1,
    edge_marker_size=15,
    max_order=None,
    node_labels=False,
    hyperedge_labels=False,
    settings=None,
    **kwargs,
):
    """Draw a directed hypergraph

    Parameters
    ----------
    DH : DirectedHypergraph
        The directed hypergraph to draw.
    ax : matplotlib.pyplot.axes, optional
        Axis to draw on. If None (default), get the current axes.
    lines_fc : str, dict, iterable, optional
        Color of the hyperedges (lines). If str, use the same color for all hyperedges.
        If a dict, must contain (hyperedge_id: color_str) pairs. If other iterable,
        assume the colors are specified in the same order as the hyperedges are found
        in DH.edges. If None (default), use the size of the hyperedges.
    lines_lw : int, float, dict, iterable, optional
        Line width of the hyperedges (lines). If int or float, use the same width for
        all hyperedges. If a dict, must contain (hyperedge_id: width) pairs. If other
        iterable, assume the widths are specified in the same order as the hyperedges
        are found in DH.edges. By default, 1.5.
    line_head_width : float, optional
        Length of arrows' heads. By default, 0.05
    node_fc : str, dict, iterable, or NodeStat, optional
        Color of the nodes.  If str, use the same color for all nodes.  If a dict, must
        contain (node_id: color_str) pairs.  If other iterable, assume the colors are
        specified in the same order as the nodes are found in H.nodes. If NodeStat, use
        the colormap specified with node_fc_cmap. By default, "white".
    node_ec : str, dict, iterable, or NodeStat, optional
        Color of node borders.  If str, use the same color for all nodes.  If a dict,
        must contain (node_id: color_str) pairs.  If other iterable, assume the colors
        are specified in the same order as the nodes are found in H.nodes. If NodeStat,
        use the colormap specified with node_ec_cmap. By default, "black".
    node_lw : int, float, dict, iterable, or NodeStat, optional
        Line width of the node borders in pixels.  If int or float, use the same width
        for all node borders.  If a dict, must contain (node_id: width) pairs.  If
        iterable, assume the widths are specified in the same order as the nodes are
        found in H.nodes. If NodeStat, use a monotonic linear interpolation defined
        between min_node_lw and max_node_lw. By default, 1.
    node_size : int, float, dict, iterable, or NodeStat, optional
        Radius of the nodes in pixels.  If int or float, use the same radius for all
        nodes.  If a dict, must contain (node_id: radius) pairs.  If iterable, assume
        the radiuses are specified in the same order as the nodes are found in
        H.nodes. If NodeStat, use a monotonic linear interpolation defined between
        min_node_size and max_node_size. By default, 15.
    edge_marker_toggle: bool, optional
        If True then marker representing the hyperedges are drawn. By default True.
    edge_marker_fc: str, dict, iterable, optional
        Filling color of the hyperedges (markers). If str, use the same color for all hyperedges.
        If a dict, must contain (hyperedge_id: color_str) pairs. If other iterable,
        assume the colors are specified in the same order as the hyperedges are found
        in DH.edges. If None (default), use the size of the hyperedges.
    edge_marker_ec: str, dict, iterable, optional
        Edge color of the hyperedges (markers). If str, use the same color for all hyperedges.
        If a dict, must contain (hyperedge_id: color_str) pairs. If other iterable,
        assume the colors are specified in the same order as the hyperedges are found
        in DH.edges. If None (default), use the size of the hyperedges.
    edge_marker: str, optional
        Marker used for the hyperedges. By default 's' (square marker).
    max_order : int, optional
        Maximum of hyperedges to plot. If None (default), plots all orders.
    node_labels : bool or dict, optional
        If True, draw ids on the nodes. If a dict, must contain (node_id: label) pairs.
        By default, False.
    hyperedge_labels : bool or dict, optional
        If True, draw ids on the hyperedges. If a dict, must contain (edge_id: label)
        pairs.  By default, False.
    **kwargs : optional args
        Alternate default values. Values that can be overwritten are the following:
        * min_node_size
        * max_node_size
        * min_node_lw
        * max_node_lw
        * node_fc_cmap
        * node_ec_cmap
        * min_lines_lw
        * max_lines_lw
        * lines_fc_cmap
        * edge_fc_cmap
        * edge_marker_fc_cmap
        * edge_marker_ec_cmap

    Returns
    -------
    ax : matplotlib.pyplot.axes

    Raises
    ------
    XGIError
        If something different than a DiHypergraph is passed.

    See Also
    --------
    draw
    draw_nodes
    draw_node_labels

    """
    if not isinstance(DH, DiHypergraph):
        raise XGIError("The input must be a DiHypergraph")

    if settings is None:
        settings = {
            "min_node_size": 10.0,
            "max_node_size": 30.0,
            "min_node_lw": 1.0,
            "max_node_lw": 5.0,
            "node_fc_cmap": cm.Reds,
            "node_ec_cmap": cm.Greys,
            "min_lines_lw": 2.0,
            "max_lines_lw": 10.0,
            "lines_fc_cmap": cm.Blues,
            "edge_marker_fc_cmap": cm.Blues,
            "edge_marker_ec_cmap": cm.Greys,
        }

    settings.update(kwargs)

    if ax is None:
        ax = plt.gca()

    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    ax.axis("off")

    # convert to hypergraph in order to use the augmented projection function
    H_conv = convert.convert_to_hypergraph(DH)

    if not max_order:
        max_order = max_edge_order(H_conv)

    lines_lw = _scalar_arg_to_dict(
        lines_lw, H_conv.edges, settings["min_lines_lw"], settings["max_lines_lw"]
    )

    if lines_fc is None:
        lines_fc = H_conv.edges.size

    lines_fc = _color_arg_to_dict(lines_fc, H_conv.edges, settings["lines_fc_cmap"])

    if edge_marker_fc is None:
        edge_marker_fc = H_conv.edges.size

    edge_marker_fc = _color_arg_to_dict(
        edge_marker_fc, H_conv.edges, settings["edge_marker_fc_cmap"]
    )

    if edge_marker_ec is None:
        edge_marker_ec = H_conv.edges.size

    edge_marker_ec = _color_arg_to_dict(
        edge_marker_ec, H_conv.edges, settings["edge_marker_ec_cmap"]
    )

    node_size = _scalar_arg_to_dict(
        node_size, H_conv.nodes, settings["min_node_size"], settings["max_node_size"]
    )

    G_aug = _augmented_projection(H_conv)
    for dyad in H_conv.edges.filterby("size", 2).members():
        try:
            index = max(n for n in G_aug.nodes if isinstance(n, int)) + 1
        except ValueError:
            # The list of node-labels has no integers, so I start from 0
            index = 0
        G_aug.add_edges_from([[list(dyad)[0], index], [list(dyad)[1], index]])

    phantom_nodes = [n for n in list(G_aug.nodes) if n not in list(H_conv.nodes)]
    pos = spring_layout(G_aug)

    for id, he in DH.edges.members(dtype=dict).items():
        d = len(he) - 1
        if d > 0:
            # identify the center of the edge in the augemented projection
            center = [n for n in phantom_nodes if set(G_aug.neighbors(n)) == he][0]
            x_center, y_center = pos[center]
            for node in DH.edges.dimembers(id)[0]:
                x_coords = [pos[node][0], x_center]
                y_coords = [pos[node][1], y_center]
                line = plt.Line2D(
                    x_coords,
                    y_coords,
                    color=lines_fc[id],
                    lw=lines_lw[id],
                    zorder=max_order - d,
                )
                ax.add_line(line)
            for node in DH.edges.dimembers(id)[1]:
                dx, dy = pos[node][0] - x_center, pos[node][1] - y_center
                # the following to avoid the point of the arrow overlapping the node
                distance = np.hypot(dx, dy)
                direction_vector = np.array([dx, dy]) / distance
                shortened_distance = (
                    distance - node_size[node] * 0.003
                )  # Calculate the shortened length
                dx = direction_vector[0] * shortened_distance
                dy = direction_vector[1] * shortened_distance
                arrow = FancyArrow(
                    x_center,
                    y_center,
                    dx,
                    dy,
                    color=lines_fc[id],
                    width=lines_lw[id] * 0.001,
                    length_includes_head=True,
                    head_width=line_head_width,
                    zorder=max_order - d,
                )
                ax.add_patch(arrow)
            if edge_marker_toggle:
                ax.scatter(
                    x=x_center,
                    y=y_center,
                    marker=edge_marker,
                    s=edge_marker_size**2,
                    c=edge_marker_fc[id],
                    edgecolors=edge_marker_ec[id],
                    linewidths=edge_marker_lw,
                    zorder=max_order,
                )

    if hyperedge_labels:
        # Get all valid keywords by inspecting the signatures of draw_node_labels
        valid_label_kwds = signature(draw_hyperedge_labels).parameters.keys()
        # Remove the arguments of this function (draw_networkx)
        valid_label_kwds = valid_label_kwds - {"H", "pos", "ax", "hyperedge_labels"}
        if any([k not in valid_label_kwds for k in kwargs]):
            invalid_args = ", ".join([k for k in kwargs if k not in valid_label_kwds])
            raise ValueError(f"Received invalid argument(s): {invalid_args}")
        label_kwds = {k: v for k, v in kwargs.items() if k in valid_label_kwds}
        if "font_size_edges" not in label_kwds:
            label_kwds["font_size_edges"] = 6
        draw_hyperedge_labels(H_conv, pos, hyperedge_labels, ax_edges=ax, **label_kwds)

    draw_nodes(
        H=H_conv,
        pos=pos,
        ax=ax,
        node_fc=node_fc,
        node_ec=node_ec,
        node_lw=node_lw,
        node_size=node_size,
        zorder=max_order,
        params=settings,
        node_labels=node_labels,
        **kwargs,
    )

    # compute axis limits
    _update_lims(pos, ax)

    return ax



"""Method for converting to a line graph."""

from itertools import combinations

import networkx as nx

from ..exception import XGIError

__all__ = ["to_line_graph"]


def to_line_graph(H, s=1, weights=None):
    """The s-line graph of the hypergraph.

    The s-line graph of the hypergraph `H` is the graph whose
    nodes correspond to each hyperedge in `H`, linked together
    if they share at least s vertices.

    Optional edge weights correspond to the size of the
    intersection between the hyperedges, optionally
    normalized by the size of the smaller hyperedge.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest
    s : int
        The intersection size to consider edges
        as connected, by default 1.
    weights : str or None
        Specify whether to return a weighted line graph. If None,
        returns an unweighted line graph. If 'absolute', includes
        edge weights corresponding to the size of intersection
        between hyperedges. If 'normalized', includes edge weights
        normalized by the size of the smaller hyperedge.

    Returns
    -------
    LG : networkx.Graph
         The line graph associated to the Hypergraph

    References
    ----------
    "Hypernetwork science via high-order hypergraph walks", by Sinan G. Aksoy, Cliff
    Joslyn, Carlos Ortiz Marrero, Brenda Praggastis & Emilie Purvine.
    https://doi.org/10.1140/epjds/s13688-020-00231-0

    """
    if weights not in [None, "absolute", "normalized"]:
        raise XGIError(
            f"{weights} not a valid weights option. Choices are "
            "None, 'absolute', and 'normalized'."
        )
    LG = nx.Graph()

    edge_label_dict = {tuple(edge): index for index, edge in H._edge.items()}

    LG.add_nodes_from(H.edges)

    for edge1, edge2 in combinations(H.edges.members(), 2):
        # Check that the intersection size is larger than s
        intersection_size = len(edge1.intersection(edge2))
        if intersection_size >= s:
            if not weights:
                # Add unweighted edge
                LG.add_edge(
                    edge_label_dict[tuple(edge1)], edge_label_dict[tuple(edge2)]
                )
            else:
                # Compute the (normalized) weight
                weight = intersection_size
                if weights == "normalized":
                    weight /= min([len(edge1), len(edge2)])
                # Add edge with weight
                LG.add_edge(
                    edge_label_dict[tuple(edge1)],
                    edge_label_dict[tuple(edge2)],
                    weight=weight,
                )

    return LG



"""Methods for converting to and from a Pandas dataframe."""

from collections import defaultdict

import pandas as pd

from ..core import SimplicialComplex
from ..exception import XGIError
from ..generators import empty_hypergraph

__all__ = ["from_bipartite_pandas_dataframe", "to_bipartite_pandas_dataframe"]


def from_bipartite_pandas_dataframe(
    df, create_using=None, node_column=0, edge_column=1
):
    """Create a hypergraph from a pandas dataframe given
    specified node and edge columns.

    Parameters
    ----------
    df : Pandas dataframe
        A dataframe where specified columns list the node IDs
        and the associated edge IDs
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    node_column : hashable, optional
        The column with the node IDs, by default 0
        Can specify names or indices
    edge_column : hashable, optional
        The column with the edge IDs, by default 1
        Can specify names or indices

    Returns
    -------
    Hypergraph object
        The constructed hypergraph

    Raises
    ------
    XGIError
        Raises an error if the user specifies invalid column names
    """
    H = empty_hypergraph(create_using)

    # try to get by labels first
    try:
        d = df[[node_column, edge_column]]
    except KeyError:
        # try to index the labels
        try:
            columns = list(df.columns)
            d = df[[columns[node_column], columns[edge_column]]]
        except (KeyError, TypeError):
            raise XGIError("Invalid columns specified")

    if isinstance(H, SimplicialComplex):
        simplex_list = defaultdict(list)
        for line in d.itertuples(index=False):
            if line[0] not in simplex_list[line[1]]:
                simplex_list[line[1]].append(line[0])

        H.add_simplices_from(list(simplex_list.values()))
    else:
        for line in d.itertuples(index=False):
            node = line[0]
            edge = line[1]
            H.add_node_to_edge(edge, node)

    return H


def to_bipartite_pandas_dataframe(H):
    """Create a two column dataframe from a hypergraph.

    Parameters
    ----------
    H : Hypergraph or Simplicial Complex
        A dataframe where specified columns list the node IDs
        and the associated edge IDs

    Returns
    -------
    Pandas Dataframe object
        A two column dataframe

    Raises
    ------
    XGIError
        Raises an error if the user specifies invalid column names
    """
    data = []
    for id1, members in H._node.items():
        for id2 in members:
            data.append([id1, id2])
    return pd.DataFrame(data, columns=["Node ID", "Edge ID"])
